import itertools
import math
import re
import structlog
import uuid

from collections import OrderedDict
from copy import deepcopy
from dcicutils.misc_utils import ignored
from elasticsearch import TransportError, RequestError, ConnectionTimeout
from elasticsearch_dsl import Search
from pyramid.httpexceptions import HTTPBadRequest
from pyramid.view import view_config
from urllib.parse import urlencode
from webob.multidict import MultiDict

from ..elasticsearch import ELASTIC_SEARCH
from ..elasticsearch.create_mapping import determine_if_is_date_field
from ..elasticsearch.indexer_utils import get_namespaced_index
from ..embed import make_subrequest
from ..interfaces import TYPES, COLLECTIONS
from ..resources import AbstractCollection
from ..typeinfo import AbstractTypeInfo
from ..util import find_collection_subtypes, crawl_schema


log = structlog.getLogger(__name__)


def includeme(config):
    config.add_route('search', '/search{slash:/?}')
    config.add_route('browse', '/browse{slash:/?}')
    config.scan(__name__)


sanitize_search_string_re = re.compile(r'[\\\+\-\&\|\!\(\)\{\}\[\]\^\~\:\/\\\*\?]')


COMMON_EXCLUDED_URI_PARAMS = [
    'frame', 'format', 'limit', 'sort', 'from', 'field',
    'mode', 'redirected_from', 'datastore', 'referrer',
    'currentAction'
]


@view_config(route_name='search', request_method='GET', permission='search')
def search(context, request, search_type=None, return_generator=False, forced_type='Search', custom_aggregations=None):
    """
    Search view connects to ElasticSearch and returns the results
    """
    types = request.registry[TYPES]
    # list of item types used from the query
    doc_types = set_doc_types(request, types, search_type)
    # sets request.normalized_params
    search_base = normalize_query(request, types, doc_types)
    # == INITIALIZE RESULT ==
    result = {
        '@context': request.route_path('jsonld_context'),
        '@id': '/' + forced_type.lower() + '/' + search_base,
        '@type': [forced_type],
        'title': forced_type,
        'filters': [],
        'facets': [],
        '@graph': [],
        'notification': '',
        'sort': {}
    }
    principals = request.effective_principals
    es = request.registry[ELASTIC_SEARCH]

    from_, size = get_pagination(request)

    # get desired frame for this search
    search_frame = request.normalized_params.get('frame', 'embedded')

    # == PREPARE SEARCH TERM ==
    prepared_terms = prepare_search_term(request)

    schemas = [types[item_type].schema for item_type in doc_types]

    # set ES index based on doc_type (one type per index)
    # if doc_type is item, search all indexes by setting es_index to None
    # If multiple, search all specified
    namespaced_star = get_namespaced_index(request, '*')
    if 'Item' in doc_types:
        es_index = namespaced_star
    else:
        es_index = find_index_by_doc_types(request, doc_types, ['Item'])

    # establish elasticsearch_dsl class that will perform the search
    search = Search(using=es, index=es_index)

    # set up clear_filters path
    result['clear_filters'] = clear_filters_setup(request, doc_types, forced_type)

    # == SET TYPE FILTERS ==
    build_type_filters(result, request, doc_types, types)

    # get the fields that will be used as source for the search
    # currently, supports frame=raw/object but live faceting does not work
    # this is okay because the only non-embedded access will be programmatic
    source_fields = sorted(list_source_fields(request, doc_types, search_frame))

    # == GET FILTERED QUERY ==
    # Builds filtered query which supports multiple facet selection
    search, string_query = build_query(search, prepared_terms, source_fields)

    # == Set sort order ==
    search = set_sort_order(request, search, prepared_terms, types, doc_types, result)
    # TODO: implement BOOST here?

    # == Set filters ==
    search, query_filters = set_filters(request, search, result, principals, doc_types)

    # == Set starting facets ==
    facets = initialize_facets(request, doc_types, prepared_terms, schemas)

    # == Adding facets, plus any optional custom aggregations. ==
    # Uses 'size' and 'from_' to conditionally skip (no facets if from > 0; no aggs if size > 0).
    search = set_facets(search, facets, query_filters, string_query, request, doc_types, custom_aggregations, size, from_)

    # == Add preference from session, if available ==
    search_session_id = None
    if request.__parent__ is None and not return_generator and size != 'all':  # Probably unnecessary, but skip for non-paged, sub-reqs, etc.
        search_session_id = request.cookies.get('searchSessionID', 'SESSION-' + str(uuid.uuid1()))
        search = search.params(preference=search_session_id)

    # == Execute the query ==
    if size == 'all':
        es_results = execute_search_for_all_results(search)
    else:
        size_search = search[from_:from_ + size]
        es_results = execute_search(size_search)

    # == Record total number of hits ==
    result['total'] = total = es_results['hits']['total']['value']
    result['facets'] = format_facets(es_results, facets, total, search_frame)
    result['aggregations'] = format_extra_aggregations(es_results)

    # Add batch actions
    # TODO: figure out exactly what this does. Provide download URLs?
    # Implement later
    # result.update(search_result_actions(request, doc_types, es_results))

    # == Add all link for collections ==
    if size not in (None, 'all') and size < result['total']:
        params = [(k, v) for k, v in request.normalized_params.items() if k != 'limit']
        params.append(('limit', 'all'))
        if context:
            result['all'] = '%s?%s' % (request.resource_path(context), urlencode(params))

    # add actions (namely 'add')
    result['actions'] = get_collection_actions(request, types[doc_types[0]])

    if not result['total']:
        # http://googlewebmastercentral.blogspot.com/2014/02/faceted-navigation-best-and-5-of-worst.html
        request.response.status_code = 404
        result['notification'] = 'No results found'
        result['@graph'] = []
        return result if not return_generator else []

    columns = build_table_columns(request, schemas, doc_types)
    if columns:
        result['columns'] = columns

    result['notification'] = 'Success'

    # == Format results for JSON-LD ==
    graph = format_results(request, es_results['hits']['hits'], search_frame)

    if request.__parent__ is not None or return_generator:
        if return_generator:
            return graph
        else:
            result['@graph'] = list(graph)
            return result

    result['@graph'] = list(graph)
    if search_session_id:  # Is 'None' if e.g. limit=all
        request.response.set_cookie('searchSessionID', search_session_id)  # Save session ID for re-requests / subsequent pages.
    return result


@view_config(route_name='browse', request_method='GET', permission='search')
def browse(context, request, search_type='ExperimentSetReplicate', return_generator=False):
    """
    Simply use search results for browse view
    """
    return search(context, request, search_type, return_generator, forced_type='Browse')


@view_config(context=AbstractCollection, permission='list', request_method='GET')
def collection_view(context, request):
    """
    Simply use search results for collections views (e.g./biosamples/)
    This is a redirect directly to the search page
    """
    return search(context, request, context.type_info.name, False, forced_type='Search')


def get_collection_actions(request, type_info):
    collection = request.registry[COLLECTIONS].get(type_info.name)
    if collection and hasattr(collection, 'actions'):
        return collection.actions(request)
    else:
        return None


def get_pagination(request):
    """
    Fill from_ and size parameters for search if given in the query string
    """
    from_ = request.normalized_params.get('from', 0)
    size = request.normalized_params.get('limit', 25)
    if size in ('all', ''):
        size = "all"
    else:
        try:
            size = int(size)
        except ValueError:
            size = 25
        try:
            from_ = int(from_)
        except ValueError:
            size = 0
    return from_, size


def get_all_subsequent_results(initial_search_result, search, extra_requests_needed_count, size_increment):
    from_ = 0
    while extra_requests_needed_count > 0:
        # print(str(extra_requests_needed_count) + " requests left to get all results.")
        from_ = from_ + size_increment
        subsequent_search = search[from_:from_ + size_increment]
        subsequent_search_result = execute_search(subsequent_search)
        extra_requests_needed_count -= 1
        for hit in subsequent_search_result['hits'].get('hits', []):
            yield hit


def execute_search_for_all_results(search):
    size_increment = 100  # Decrease this to like 5 or 10 to test.

    first_search = search[0:size_increment]  # get aggregations from here
    es_result = execute_search(first_search)

    total_results_expected = es_result['hits'].get('total', 0)
    extra_requests_needed_count = int(math.ceil(total_results_expected / size_increment)) - 1  # Decrease by 1 (first es_result already happened)

    if extra_requests_needed_count > 0:
        es_result['hits']['hits'] = itertools.chain(es_result['hits']['hits'], get_all_subsequent_results(es_result, search, extra_requests_needed_count, size_increment))
    return es_result


def normalize_query(request, types, doc_types):
    """
    Normalize the query by calculating and setting request.normalized_params
    (a webob MultiDict) that is derived from custom query rules and also
    the list of doc_types specified by set_doc_types(). The normalize_param
    helper function finds field_schema for each query parameter and enforces
    a set of rules (see below). If the query item types differ from doc_types,
    override with doc_types

    Args:
        request: the current Request
        types: registry[TYPES]
        doc_types (list): item_types to use for the search

    Returns:
        string: query string built from normalized params
    """
    def normalize_param(key, val):
        """
        Process each key/val in the original query param. As part of this,
        obtain the field schema for each parameter.
        Current rules:
        - for 'type', get name from types (from the registry)
        - append '.display_title' to any terminal linkTo query field
        """
        # type param is a special case. use the name from TypeInfo
        if key == 'type' and val in types:
            return (key, types[val].name)

        # find schema for field parameter and drill down into arrays/subobjects
        field_schema = schema_for_field(key, request, doc_types)
        while field_schema and ('items' in field_schema or 'properties' in field_schema):
            try:
                field_schema = field_schema['items']
            except KeyError:
                pass
            try:
                field_schema = field_schema['properties']
            except KeyError:
                pass
        if field_schema and 'linkTo' in field_schema:
            # add display_title to terminal linkTo query fields
            if key.endswith('!'):  # handle NOT
                return (key[:-1] + '.display_title!', val)
            return (key + '.display_title', val)
        else:
            return (key, val)

    normalized_params = (
        normalize_param(k, v)
        for k, v in request.params.items()
    )
    # use a MultiDict to emulate request.params
    normalized_params = MultiDict(normalized_params)
    # overwrite 'type' if not equal to doc_types
    if set(normalized_params.getall('type')) != set(doc_types):
        if 'type' in normalized_params:
            del normalized_params['type']
        for dtype in doc_types:
            normalized_params.add('type', dtype)
    # add the normalized params to the request
    # these will be used in place of request.params for the rest of search
    setattr(request, 'normalized_params', normalized_params)
    # the query string of the normalized search
    qs = '?' + urlencode([
        (k.encode('utf-8'), v.encode('utf-8'))
        for k, v in request.normalized_params.items()
    ])
    return qs


def clear_filters_setup(request, doc_types, forced_type):
    """
    Clear Filters URI path

    Make a URI path that clears all non-datatype filters
    and leaves in `q` (search query) params, if present.
    Also preserves currentAction=selection, if is set.

    Returns:
        A URL path
    """
    seach_query_specs = request.normalized_params.getall('q')
    seach_query_url = urlencode([("q", seach_query) for seach_query in seach_query_specs])
    # types_url will always be present (always >=1 doc_type)
    types_url = urlencode([("type", typ) for typ in doc_types])
    current_action = request.normalized_params.get('currentAction')

    clear_qs = types_url or ''
    if seach_query_url:
        clear_qs += '&' + seach_query_url
    if current_action == 'selection':
        clear_qs += '&currentAction=selection'
    current_search_sort = request.normalized_params.getall('sort')
    current_search_sort_url = urlencode([("sort", s) for s in current_search_sort])
    if current_search_sort_url:
        clear_qs += '&' + current_search_sort_url
    return request.route_path(forced_type.lower(), slash='/') + (('?' + clear_qs) if clear_qs else '')


def build_type_filters(result, request, doc_types, types):
    """
    Set the type filters for the search. If no doc_types, default to Item
    """
    if not doc_types:
        doc_types = ['Item']
    else:
        for item_type in doc_types:
            ti = types[item_type]
            qs = urlencode([
                (k.encode('utf-8'), v.encode('utf-8'))
                for k, v in request.normalized_params.items() if not (k == 'type' and types.all.get('Item' if v == '*' else v) is ti)
            ])
            result['filters'].append({
                'field': 'type',
                'term': ti.name,
                'remove': '{}?{}'.format(request.path, qs)
            })


def prepare_search_term(request):
    """
    Prepares search terms by making a dictionary where the keys are fields
    and the values are arrays of query strings
    Ignore certain keywords, such as type, format, and field
    """
    prepared_terms = {}
    # prepared_vals = []
    # In case it helps, I think request.normalized_params is a MultiDict. -kmp 7-Aug-2022
    for field, val in request.normalized_params.items():  # was .iteritems(), but that went away in Python 3
        if field.startswith('validation_errors') or field.startswith('aggregated_items'):
            continue
        elif field == 'q':  # searched string has field 'q'
            # people shouldn't provide multiple queries, but if they do,
            # combine them with AND logic
            if 'q' in prepared_terms:
                join_list = [prepared_terms['q'], val]
                prepared_terms['q'] = ' AND '.join(join_list)
            else:
                prepared_terms['q'] = val
        elif field not in COMMON_EXCLUDED_URI_PARAMS + ['type']:
            if 'embedded.' + field not in prepared_terms.keys():
                prepared_terms['embedded.' + field] = []
            prepared_terms['embedded.' + field].append(val)
    return prepared_terms


def set_doc_types(request, types, search_type):
    """
    Set the type of documents resulting from the search; order and check for
    invalid types as well. If a forced search_type is enforced, use that;
    otherwise, set types from the query params. Default to Item if none set.

    Args:
        request: the current Request
        types: registry[TYPES]
        search_type (str): forced search item type

    Returns:
        list: the string item types to use for the search

    Raises:
        HTTPBadRequest: if an invalid item type is supplied
    """
    doc_types = []
    if search_type is None:
        doc_types = request.params.getall('type')
        if '*' in doc_types:
            doc_types = ['Item']
    else:
        doc_types = [search_type]
    # Normalize to item_type
    try:
        doc_types = sorted({types[name].name for name in doc_types})
    except KeyError:
        # Check for invalid types
        bad_types = [t for t in doc_types if t not in types]
        msg = "Invalid type: {}".format(', '.join(bad_types))
        raise HTTPBadRequest(explanation=msg)
    if len(doc_types) == 0:
        doc_types = ['Item']
    return doc_types


def get_search_fields(request, doc_types):
    """
    Returns set of columns that are being searched and highlights
    """
    fields = {'uuid'}
    highlights = {}
    types = request.registry[TYPES]
    for doc_type in doc_types:
        type_info = types[doc_type]
        for value in type_info.schema.get('boost_values', ()):
            fields.add('embedded.' + value)
            highlights['embedded.' + value] = {}
    return fields, highlights


def list_source_fields(request, doc_types, frame):
    """
    Returns set of fields that are requested by user or default fields.
    These fields are used to further limit the results from the search.
    Note that you must provide the full fieldname with embeds, such as:
    'field=biosample.biosource.individual.organism.name' and not just
    'field=name'
    """
    fields_requested = request.normalized_params.getall('field')
    if fields_requested:
        fields = ['embedded.@id', 'embedded.@type']
        for field in fields_requested:
            fields.append('embedded.' + field)
    elif frame in ['embedded', 'object', 'raw']:
        if frame != 'embedded':
            # frame=raw corresponds to 'properties' in ES
            if frame == 'raw':
                frame = 'properties'
            # let embedded be searched as well (for faceting)
            fields = ['embedded.*', frame + '.*']
        else:
            fields = [frame + '.*']
    else:
        fields = ['embedded.*']
    return fields


def build_query(search, prepared_terms, source_fields):
    """
    Prepare the query within the Search object.
    """
    query_info = {}
    string_query = None
    # set _source fields for the search
    search = search.source(list(source_fields))
    # prepare the query from prepared_terms
    for field, value in prepared_terms.items():
        if field == 'q':
            query_info['query'] = value
            query_info['lenient'] = True
            query_info['default_operator'] = 'AND'
            query_info['fields'] = ['_all']
            break
    if query_info != {}:
        string_query = {'must': {'simple_query_string': query_info}}
        query_dict = {'query': {'bool': string_query}}
    else:
        query_dict = {'query': {'bool': {}}}
    search.update_from_dict(query_dict)
    return search, string_query


def set_sort_order(request, search, search_term, types, doc_types, result):
    """
    sets sort order for elasticsearch results
    example: /search/?type=Biosource&sort=display_title
    will sort by display_title in ascending order. To set descending order,
    use the "-" flag: sort_by=-date_created.
    Sorting is done alphatbetically, case sensitive by default.
    TODO: add a schema flag for case sensitivity/insensitivity?

    ES5: simply pass in the sort OrderedDict into search.sort
    """
    sort = OrderedDict()
    result_sort = OrderedDict()
    if len(doc_types) == 1:
        type_schema = types[doc_types[0]].schema
    else:
        type_schema = None

    def add_to_sort_dict(requested_sort):
        if requested_sort.startswith('-'):
            name = requested_sort[1:]
            order = 'desc'
        else:
            name = requested_sort
            order = 'asc'
        sort_schema = type_schema.get('properties', {}).get(name) if type_schema else None
        if sort_schema:
            sort_type = sort_schema.get('type')
        else:
            sort_type = 'string'

        # ES type != schema types
        if sort_type == 'integer':
            sort['embedded.' + name] = result_sort[name] = {
                'order': order,
                'unmapped_type': 'long',
                'missing': '_last'
            }
        elif sort_type == 'number':
            sort['embedded.' + name] = result_sort[name] = {
                'order': order,
                'unmapped_type': 'float',
                'missing': '_last'
            }
        else:
            # fallback case, applies to all string type:string fields
            sort['embedded.' + name + '.lower_case_sort'] = result_sort[name] = {
                'order': order,
                'unmapped_type': 'keyword',
                'missing': '_last'
            }

    # Prefer sort order specified in request, if any
    requested_sorts = request.normalized_params.getall('sort')
    if requested_sorts:
        for rs in requested_sorts:
            add_to_sort_dict(rs)

    text_search = search_term.get('q')

    # Otherwise we use a default sort only when there's no text search to be ranked
    if not sort and (text_search == '*' or not text_search):
        # If searching for a single type, look for sort options in its schema
        if type_schema:
            if 'sort_by' in type_schema:
                for k, v in type_schema['sort_by'].items():
                    # Should always sort on raw field rather than analyzed field
                    # OR search on lower_case_sort for case insensitive results
                    sort['embedded.' + k + '.lower_case_sort'] = result_sort[k] = v
        # Default is most recent first, then alphabetical by label
        if not sort:
            sort['embedded.date_created.raw'] = result_sort['date_created'] = {
                'order': 'desc',
                'unmapped_type': 'keyword',
            }
            sort['embedded.label.raw'] = result_sort['label'] = {
                'order': 'asc',
                'missing': '_last',
                'unmapped_type': 'keyword',
            }
    elif not sort and text_search and text_search != '*':
        search = search.sort(                   # Multi-level sort. See http://www.elastic.co/guide/en/elasticsearch/guide/current/_sorting.html#_multilevel_sorting & https://stackoverflow.com/questions/46458803/python-elasticsearch-dsl-sorting-with-multiple-fields
            {'_score': {"order": "desc"}},
            {'embedded.date_created.raw': {'order': 'desc', 'unmapped_type': 'keyword'},
             'embedded.label.raw': {'order': 'asc', 'unmapped_type': 'keyword', 'missing': '_last'}},
            {'_uid': {'order': 'asc'}}  # 'embedded.uuid.raw' (instd of _uid) sometimes results in 400 bad request : 'org.elasticsearch.index.query.QueryShardException: No mapping found for [embedded.uuid.raw] in order to sort on'
        )
        result['sort'] = result_sort = {'_score': {"order": "desc"}}
        return search

    if sort and result_sort:
        result['sort'] = result_sort
        search = search.sort(sort)
    return search


def set_filters(request, search, result, principals, doc_types):
    """
    Sets filters in the query
    """

    # these next two dictionaries should each have keys equal to query_field
    # and values: must_terms: [<list of terms>], must_not_terms: [<list of terms>], add_no_value: True/False/None
    field_filters = {
        'principals_allowed.view': {
            'must_terms': principals,
            'must_not_terms': [],
            'add_no_value': None
        },
        'embedded.@type.raw': {
            'must_terms': doc_types,
            'must_not_terms': [],
            'add_no_value': None
        },
        'embedded.status.raw': {
            'must_terms': [],
            'must_not_terms': [],
            'add_no_value': None
        }
    }

    range_filters = {}

    # Exclude status=deleted Items unless explicitly requested/filtered-in.
    if 'deleted' not in request.normalized_params.getall('status'):
        field_filters['embedded.status.raw']['must_not_terms'].append('deleted')
    if 'replaced' not in request.normalized_params.getall('status'):
        field_filters['embedded.status.raw']['must_not_terms'].append('replaced')

    # Exclude type=TrackingItem and type=OntologyTerm from results unless are explictly specified
    if 'TrackingItem' not in doc_types:
        field_filters['embedded.@type.raw']['must_not_terms'].append('TrackingItem')
    if 'OntologyTerm' not in doc_types:
        field_filters['embedded.@type.raw']['must_not_terms'].append('OntologyTerm')

    for field, term in request.normalized_params.items():
        not_field = False  # keep track if query is NOT (!)
        exists_field = False  # keep track of null values
        range_type = False  # If we determine is a range request (field.to, field.from), will be populated with string 'date' or 'numerical'
        range_direction = None
        if field in COMMON_EXCLUDED_URI_PARAMS + ['q']:
            continue
        elif field == 'type' and term != 'Item':
            continue
        elif term == 'No value':
            exists_field = True

        # Check for date or numerical range filters
        if (len(field) > 3 and field[-3:] == '.to') or (len(field) > 5 and field[-5:] == '.from'):
            if field[-3:] == '.to':
                f_field = field[:-3]
                range_direction = "lte"
            else:
                f_field = field[:-5]
                range_direction = "gte"

            # If schema for field is not found (and range_type thus not set),
            # then treated as ordinary term filter (likely will get 0 results)
            field_schema = schema_for_field(f_field, request, doc_types)
            if field_schema:
                range_type = 'date' if determine_if_is_date_field(f_field, field_schema) else 'numerical'

        # Add filter to result
        qs = urlencode([
            (k.encode('utf-8'), v.encode('utf-8'))
            for k, v in request.normalized_params.items()
            if (k != field or v != term)
        ])
        remove_path = '{}?{}'.format(request.path, qs)

        # default to searching type=Item rather than empty filter path
        if remove_path[-1] == '?':
            remove_path += 'type=Item'

        result['filters'].append({
            'field': field,
            'term': term,
            'remove': remove_path
        })

        # handle NOT
        if field.endswith('!'):
            field = field[:-1]
            not_field = True

        # Add filter to query
        if range_type and f_field and range_type in ('date', 'numerical'):
            query_field = 'embedded.' + f_field
        elif field.startswith('validation_errors') or field.startswith('aggregated_items'):
            query_field = field + '.raw'
        elif field == 'type':
            query_field = 'embedded.@type.raw'
        else:
            query_field = 'embedded.' + field + '.raw'

        if range_type:
            if query_field not in range_filters:
                range_filters[query_field] = {}
                if range_type == 'date':
                    range_filters[query_field]['format'] = 'yyyy-MM-dd HH:mm'

            if range_direction in ('gt', 'gte', 'lt', 'lte'):
                if len(term) == 10:
                    # Correct term to have hours, e.g. 00:00 or 23:59, if not otherwise supplied.
                    if range_direction == 'gt' or range_direction == 'lte':
                        term += ' 23:59'
                    elif range_direction == 'gte' or range_direction == 'lt':
                        term += ' 00:00'

                if range_filters[query_field].get(range_direction) is None:
                    range_filters[query_field][range_direction] = term
                else:
                    # If have a value already (e.g. multiple ranges selected), choose the widening option.
                    if range_direction == 'gt' or range_direction == 'gte':
                        if term < range_filters[query_field][range_direction]:
                            range_filters[query_field][range_direction] = term
                    elif range_direction == 'lt' or range_direction == 'lte':
                        if term > range_filters[query_field][range_direction]:
                            range_filters[query_field][range_direction] = term
        else:
            if query_field not in field_filters:
                field_filters[query_field] = {
                    'must_terms': [],
                    'must_not_terms': [],
                    'add_no_value': None
                }

            # handle case of filtering for null values
            if exists_field:
                # the value below is True when we want to include 'No value' as a filter
                field_filters[query_field]['add_no_value'] = False if not_field else True
                continue

            if not_field:
                field_filters[query_field]['must_not_terms'].append(term)
            else:
                field_filters[query_field]['must_terms'].append(term)

    must_filters = []
    must_not_filters = []
    for query_field, filters in field_filters.items():
        must_terms = {'terms': {query_field: filters['must_terms']}} if filters['must_terms'] else {}
        must_not_terms = {'terms': {query_field: filters['must_not_terms']}} if filters['must_not_terms'] else {}
        if filters['add_no_value'] is True:
            # add to must_not in an OR case, which is equivalent to filtering on 'No value'
            should_arr = [must_terms] if must_terms else []
            should_arr.append({'bool': {'must_not': {'exists': {'field': query_field}}}})
            must_filters.append({'bool': {'should': should_arr}})
        elif filters['add_no_value'] is False:
            # add to must_not in an OR case, which is equivalent to filtering on '! No value'
            should_arr = [must_terms] if must_terms else []
            should_arr.append({'exists': {'field': query_field}})
            must_filters.append({'bool': {'should': should_arr}})
        else:  # no filtering on 'No value'
            if must_terms:
                must_filters.append(must_terms)
        if must_not_terms:
            must_not_filters.append(must_not_terms)

    # lastly, add range limits to filters if given
    for range_field, range_def in range_filters.items():
        must_filters.append({
            'range': {range_field: range_def}
        })

    # To modify filters of elasticsearch_dsl Search, must call to_dict(),
    # modify that, then update from the new dict
    prev_search = search.to_dict()
    # initialize filter hierarchy
    final_filters = {'bool': {'must': must_filters, 'must_not': must_not_filters}}
    prev_search['query']['bool']['filter'] = final_filters
    search.update_from_dict(prev_search)

    return search, final_filters


def initialize_facets(request, doc_types, prepared_terms, schemas):
    """
    Initialize the facets used for the search. If searching across multiple
    doc_types, only use the default 'Data Type' and 'Status' facets.
    Add facets for custom url filters whether or not they're in the schema

    Args:
        doc_types (list): Item types (@type) for which we are performing a search for.
        prepared_terms (dict): terms to match in ES, keyed by ES field name.
        schemas (list): List of OrderedDicts of schemas for doc_types.

    Returns:
        list: tuples containing (0) ElasticSearch-formatted field name (e.g. `embedded.status`) and (1) list of terms for it.
    """

    facets = [
        # More facets will be appended to this list from item schema plus from any currently-active filters (as requested in URI params).
        ('type', {'title': 'Data Type'})
    ]
    append_facets = [
        # Facets which will be appended after those which are in & added to `facets`
        ('status', {'title': 'Status'}),

        # TODO: Re-enable below line if/when 'range' URI param queries for date & numerical fields are implemented.
        # ('date_created', {'title': 'Date Created', 'hide_from_view' : True, 'aggregation_type' : 'date_histogram' })
    ]
    validation_error_facets = [
        ('validation_errors.name', {'title': 'Validation Errors', 'order': 999})
    ]
    # hold disabled facets from schema; we also want to remove these from the prepared_terms facets
    disabled_facets = []

    # Add facets from schema if one Item type is defined.
    # Also, conditionally add extra appendable facets if relevant for type from schema.
    if len(doc_types) == 1 and doc_types[0] != 'Item':
        current_type_schema = request.registry[TYPES][doc_types[0]].schema
        if 'facets' in current_type_schema:
            schema_facets = OrderedDict(current_type_schema['facets'])
            for schema_facet in schema_facets.items():
                if schema_facet[1].get('disabled', False):
                    disabled_facets.append(schema_facet[0])
                    continue  # Skip disabled facets.
                facets.append(schema_facet)

    # == Add facets for any non-schema ?field=value filters requested in the search (unless already set) ==
    used_facets = [facet[0] for facet in facets + append_facets]
    used_facet_titles = [facet[1]['title']
                         for facet in facets + append_facets
                         if 'title' in facet[1]]
    for field in prepared_terms:
        if field.startswith('embedded'):
            split_field = field.strip().split('.')  # Will become, e.g. ['embedded', 'experiments_in_set', 'files', 'file_size', 'from']
            use_field = '.'.join(split_field[1:])

            # 'terms' is the default per-term bucket aggregation for all non-schema facets
            aggregation_type = 'terms'

            # Use the last part of the split field to get the field title
            title_field = split_field[-1]

            # if searching for a display_title, use the title of parent object
            # use `is_object_title` to keep track of this
            if title_field == 'display_title' and len(split_field) > 1:
                title_field = split_field[-2]
                is_object_title = True
            else:
                is_object_title = False

            if title_field in used_facets or title_field in disabled_facets:
                # Cancel if already in facets or is disabled
                continue

            # If we have a range filter in the URL,
            if title_field == 'from' or title_field == 'to':
                if len(split_field) == 3:
                    f_field = split_field[-2]
                    field_schema = schema_for_field(f_field, request, doc_types)
                    if field_schema:
                        title_field = f_field
                        use_field = '.'.join(split_field[1:-1])
                        aggregation_type = 'stats'

            for schema in schemas:
                if title_field in schema['properties']:
                    title_field = schema['properties'][title_field].get('title', title_field)
                    # see if the title field conflicts for is_object_title facets
                    if is_object_title and title_field in used_facet_titles:
                        title_field += ' (Title)'
                    break

            facet_tuple = (use_field, {'title': title_field, 'aggregation_type': aggregation_type})

            # At moment is equivalent to `if aggregation_type == 'stats'`` until/unless more agg types are added for _facets_.
            if aggregation_type != 'terms':
                facet_tuple[1]['hide_from_view'] = True  # Temporary until we handle these better on front-end.
                # Facet would be otherwise added twice if both `.from` and `.to` are requested.
                if facet_tuple in facets:
                    continue

            facets.append(facet_tuple)

    # Append additional facets (status, validation_errors, ...) at the end of
    # list unless were already added via schemas, etc.
    used_facets = [facet[0] for facet in facets]  # Reset this var
    for ap_facet in append_facets + validation_error_facets:
        if ap_facet[0] not in used_facets:
            facets.append(ap_facet)
        else:  # Update with better title if not already defined from e.g. requested filters.
            existing_facet_index = used_facets.index(ap_facet[0])
            if facets[existing_facet_index][1].get('title') in (None, facets[existing_facet_index][0]):
                facets[existing_facet_index][1]['title'] = ap_facet[1]['title']

    return facets


def schema_for_field(field, request, doc_types, should_log=False):
    """
    Find the schema for the given field (in embedded '.' format). Uses
    ff_utils.crawl_schema from snovault and logs any cases where there is an
    error finding the field from the schema

    Args:
        field (string): embedded field path, separated by '.'
        request: current Request object
        doc_types (list): @types for the search
        should_log (bool): logging will only occur if set to True

    Returns:
        Dictionary schema for the field, or None if not found
    """
    types = request.registry[TYPES]
    schemas = [types[dt].schema
               for dt in doc_types]

    # We cannot hash dict by list (of doc_types) so we convert to unique ordered string
    doc_type_string = ','.join(doc_types)

    cache = getattr(request, '_field_schema_cache', {})
    if (field, doc_type_string) in cache:
        return cache[(field, doc_type_string)]

    field_schema = None

    # for 'validation_errors.*' and 'aggregated_items.*',
    # schema will never be found and logging isn't helpful
    if (schemas
            and not field.startswith('validation_errors.')
            and not field.startswith('aggregated_items.')):
        # 'type' field is really '@type' in the schema
        use_field = '@type' if field == 'type' else field
        # eliminate '!' from not fields
        use_field = use_field[:-1] if use_field.endswith('!') else use_field
        for schema in schemas:
            try:
                field_schema = crawl_schema(types, use_field, schema)
            except Exception as exc:  # cannot find schema. Log and Return None
                if should_log:
                    log.warning(f'Cannot find schema in search.py. Type: {doc_types[0]}. Field: {field}',
                                field=field, error=str(exc))
            else:
                if field_schema is not None:
                    break

    # Cache result, even if not found, for this request.
    cache[(field, doc_type_string)] = field_schema
    if not hasattr(request, '_field_schema_cache'):
        setattr(request, '_field_schema_cache', cache)

    return field_schema


def is_linkto_or_object_array_root_field(field, types, doc_types):
    """Not used currently. May be useful for if we want to enabled "type" : "nested" mappings on lists of dictionaries"""
    schema = types[doc_types[0]].schema
    field_root = field.split('.')[0]
    fr_schema = (schema and schema.get('properties', {}).get(field_root, None)) or None
    if fr_schema and fr_schema['type'] == 'array' and (fr_schema['items'].get('linkTo') is not None or fr_schema['items']['type'] == 'object'):
        return True
    return False


def generate_filters_for_terms_agg_from_search_filters(query_field, search_filters, string_query):
    """
    We add a copy of our filters to each facet, minus that of
    facet's field itself so that we can get term counts for other terms filters.
    And be able to filter w/ it.

    Remove filters from fields they apply to.
    For example, the 'biosource_type' aggs should not have any
    biosource_type filter in place.
    Handle 'must' and 'must_not' filters separately

    Returns
        Copy of search_filters, minus filter for current query_field (if one set).
    """

    facet_filters = deepcopy(search_filters['bool'])

    for filter_type in ['must', 'must_not']:
        if not search_filters['bool'][filter_type]:
            continue
        for active_filter in search_filters['bool'][filter_type]:  # active_filter => e.g. { 'terms' : { 'embedded.@type.raw': ['ExperimentSetReplicate'] } }
            if 'bool' in active_filter and 'should' in active_filter['bool']:
                # handle No value case
                inner_bool = None
                inner_should = active_filter.get('bool').get('should', [])
                for or_term in inner_should:
                    # this may be naive, but assume first non-terms
                    # filter is the No value quqery
                    if 'terms' in or_term:
                        continue
                    else:
                        inner_bool = or_term
                        break
                if 'exists' in inner_bool:
                    compare_field = inner_bool['exists'].get('field')
                else:
                    # attempt to get the field from the alternative No value syntax
                    compare_field = inner_bool.get('bool', {}).get('must_not', {}).get('exists', {}).get('field')
                if compare_field == query_field and query_field != 'embedded.@type.raw':
                    facet_filters[filter_type].remove(active_filter)

            if 'terms' in active_filter:
                # there should only be one key here
                for compare_field in active_filter['terms'].keys():
                    # remove filter for a given field for that facet
                    # skip this for type facet (field = 'type')
                    # since we always want to include that filter.
                    if compare_field == query_field and query_field != 'embedded.@type.raw':
                        facet_filters[filter_type].remove(active_filter)

            elif 'range' in active_filter:
                for compare_field in active_filter['range'].keys():
                    # Do same as for terms
                    if compare_field == query_field:
                        facet_filters[filter_type].remove(active_filter)

    # add the string_query, if present, to the bool term with facet_filters
    if string_query and string_query['must']:
        # combine statements within 'must' for each
        facet_filters['must'].append(string_query['must'])

    return facet_filters


def set_facets(search, facets, search_filters, string_query, request, doc_types, custom_aggregations=None,
               size=25, from_=0):
    """
    Sets facets in the query as ElasticSearch aggregations, with each aggregation to be
    filtered by search_filters minus filter affecting facet field in order to get counts
    for other facet term options.
    ES5 - simply sets aggs by calling update_from_dict after adding them in

        :param facets:         Facet field (0) in object dot notation, and a dict or OrderedDict with title property (1).
        :type  facets:         List of tuples.
        :param search_filters: Dict of filters which are set for the ES query in set_filters
        :param string_query:   Dict holding the query_string used in the search
    """

    if from_ != 0:
        return search

    aggs = OrderedDict()

    for field, facet in facets:  # E.g. 'type','experimentset_type','experiments_in_set.award.project', ...
        field_schema = schema_for_field(field, request, doc_types, should_log=True)
        is_date_field = field_schema and determine_if_is_date_field(field, field_schema)
        is_numerical_field = field_schema and field_schema['type'] in ("integer", "float", "number")

        if field == 'type':
            query_field = 'embedded.@type.raw'
        elif field.startswith('validation_errors') or field.startswith('aggregated_items'):
            query_field = field + '.raw'
        elif facet.get('aggregation_type') in ('stats', 'date_histogram', 'histogram', 'range'):
            query_field = 'embedded.' + field
        else:
            query_field = 'embedded.' + field + '.raw'

        # Create the aggregation itself, extend facet with info to pass down to front-end
        agg_name = field.replace('.', '-')

        if facet.get('aggregation_type') == 'stats':

            if is_date_field:
                facet['field_type'] = 'date'
            elif is_numerical_field:
                facet['field_type'] = 'number'

            aggs[facet['aggregation_type'] + ":" + agg_name] = {
                'aggs': {
                    "primary_agg": {
                        'stats': {
                            'field': query_field
                        }
                    }
                },
                'filter': search_filters
            }

        else:  # Default -- facetable terms

            facet['aggregation_type'] = 'terms'
            facet_filters = generate_filters_for_terms_agg_from_search_filters(query_field, search_filters, string_query)
            term_aggregation = {
                "terms": {
                    'size': 100,            # Maximum terms returned (default=10); see https://github.com/10up/ElasticPress/wiki/Working-with-Aggregations
                    'field': query_field,
                    'missing': facet.get("missing_value_replacement", "No value")
                }
            }

            aggs[facet['aggregation_type'] + ":" + agg_name] = {
                'aggs': {
                    "primary_agg": term_aggregation
                },
                'filter': {'bool': facet_filters},
            }

        # Update facet with title, description from field_schema, if missing.
        if facet.get('title') is None and field_schema and 'title' in field_schema:
            facet['title'] = field_schema['title']
        if facet.get('description') is None and field_schema and 'description' in field_schema:
            facet['description'] = field_schema['description']

    # to achieve OR behavior within facets, search among GLOBAL results,
    # not just returned ones. to do this, wrap aggs in ['all_items']
    # and add "global": {} to top level aggs query
    # see elasticsearch global aggs for documentation (should be ES5 compliant)
    search_as_dict = search.to_dict()
    search_as_dict['aggs'] = {
        'all_items': {
            'global': {},
            'aggs': aggs
        }
    }

    if size == 0:
        # Only perform aggs if size==0 requested, to improve performance for search page queries.
        # We do currently have (hidden) monthly date histogram facets which may yet to be utilized for common size!=0 agg use cases.
        set_additional_aggregations(search_as_dict, request, doc_types, custom_aggregations)

    search.update_from_dict(search_as_dict)
    return search


def set_additional_aggregations(search_as_dict, request, doc_types, extra_aggregations=None):
    """
    Per-type aggregations may be defined in schemas. Apply them OUTSIDE of globals so they act on our current search filters.
    Warning: `search_as_dict` is modified IN PLACE.
    """

    types = request.registry[TYPES]
    schema = types[doc_types[0]].schema

    if schema.get('aggregations'):
        for schema_agg_name in schema['aggregations'].keys():
            if schema_agg_name == 'all_items':
                raise Exception('all_items is a reserved agg name and not allowed as an extra aggregation name.')
            search_as_dict['aggs'][schema_agg_name] = schema['aggregations'][schema_agg_name]

    if extra_aggregations:
        for extra_agg_name in extra_aggregations.keys():
            if extra_agg_name == 'all_items':
                raise Exception('all_items is a reserved agg name and not allowed as an extra aggregation name.')
            search_as_dict['aggs'][extra_agg_name] = extra_aggregations[extra_agg_name]

    return search_as_dict


def execute_search(search):
    """
    Execute the given Elasticsearch-dsl search. Raise HTTPBadRequest for any
    exceptions that arise.
    Args:
        search: the Elasticsearch-dsl prepared in the search() function
    Returns:
        Dictionary search results
    """
    err_exp = None
    try:
        es_results = search.execute().to_dict()
    except ConnectionTimeout as exc:
        ignored(exc)
        err_exp = 'The search failed due to a timeout. Please try a different query.'
    except RequestError as exc:
        # try to get a specific error message. May fail in some cases
        try:
            err_detail = str(exc.info['error']['root_cause'][0]['reason'])
        except Exception:
            err_detail = str(exc)
        err_exp = 'The search failed due to a request error: ' + err_detail
    except TransportError as exc:
        # most general exception
        exc_status = getattr(exc, 'status_code')
        if exc_status == 'TIMEOUT':
            err_exp = 'The search failed due to a timeout. Please try a different query.'
        else:
            err_exp = 'The search failed due to a transport error: ' + str(exc)
    except Exception as exc:
        ignored(exc)
        err_exp = 'The search failed. The DCIC team has been notified.'
    if err_exp:
        raise HTTPBadRequest(explanation=err_exp)
    return es_results  # noQA - PyCharm wrongly worries this might not have been set.


def format_facets(es_results, facets, total, search_frame='embedded'):
    """
    Format the facets for the final results based on the es results.
    Sort based off of the 'order' of the facets
    These are stored within 'aggregations' of the result.

    If the frame for the search != embedded, return no facets
    """
    ignored(total)
    result = []
    if search_frame != 'embedded':
        return result

    # Loading facets in to the results
    if 'aggregations' not in es_results:
        return result

    aggregations = es_results['aggregations']['all_items']
    used_facets = set()

    # Sort facets by order (ascending).
    # If no order is provided, assume 0 to
    # retain order of non-explicitly ordered facets
    for field, facet in sorted(facets, key=lambda fct: fct[1].get('order', 0)):
        result_facet = {
            'field': field,
            'title': facet.get('title', field),
            'total': 0
            # To be added depending on facet['aggregation_type']: 'terms', 'min', 'max', 'min_as_string', 'max_as_string', ...
        }

        result_facet.update({k: v
                             for k, v in facet.items()
                             if k not in result_facet.keys()})
        used_facets.add(field)
        field_agg_name = field.replace('.', '-')
        full_agg_name = facet['aggregation_type'] + ':' + field_agg_name

        if full_agg_name in aggregations:
            if facet['aggregation_type'] == 'stats':
                result_facet['total'] = aggregations[full_agg_name]['doc_count']
                # Used for fields on which can do range filter on, to provide min + max bounds
                for k in aggregations[full_agg_name]["primary_agg"].keys():
                    result_facet[k] = aggregations[full_agg_name]["primary_agg"][k]
            else:  # 'terms' assumed.
                # Default - terms, range, or histogram buckets. Buckets may not be present
                result_facet['terms'] = aggregations[full_agg_name]["primary_agg"]["buckets"]
                # Choosing to show facets with one term for summary info on search it provides
                if len(result_facet.get('terms', [])) < 1:
                    continue

            if len(aggregations[full_agg_name].keys()) > 2:
                result_facet['extra_aggs'] = {k: v
                                              for k, v in aggregations[field_agg_name].items()
                                              if k not in ('doc_count', "primary_agg")}

        result.append(result_facet)

    return result


def format_extra_aggregations(es_results):
    if 'aggregations' not in es_results:
        return {}
    return {k: v
            for k, v in es_results['aggregations'].items()
            if k != 'all_items'}


def format_results(request, hits, search_frame):
    """
    Loads results to pass onto UI
    Will retrieve the desired frame from the search hits and automatically
    add 'validation_errors' and 'aggregated_items' frames if they are present
    """
    fields_requested = request.normalized_params.getall('field')
    if fields_requested:
        frame = 'embedded'
    elif search_frame:
        frame = search_frame
    else:
        frame = 'embedded'

    if frame in ['embedded', 'object', 'raw']:
        # transform 'raw' to 'properties', which is what is stored in ES
        if frame == 'raw':
            frame = 'properties'
        for hit in hits:
            frame_result = hit['_source'][frame]
            if 'validation_errors' in hit['_source'] and 'validation_errors' not in frame_result:
                frame_result['validation_errors'] = hit['_source']['validation_errors']
            if 'aggregated_items' in hit['_source'] and 'aggregated_items' not in frame_result:
                frame_result['aggregated_items'] = hit['_source']['aggregated_items']
            yield frame_result
        return


def find_index_by_doc_types(request, doc_types, ignore):
    """
    Find the correct index(es) to be search given a list of doc_types.
    The types in doc_types are the item class names, formatted like
    'Experiment HiC' and index names are the item types, formatted like
    'experiment_hi_c'.
    Ignore any collection names provided in the ignore param, an array.
    Formats output indexes as a string usable by elasticsearch
    """
    indexes = []
    for doc_type in doc_types:
        if doc_type in ignore:
            continue
        else:
            result = find_collection_subtypes(request.registry, doc_type)
            namespaced_results = map(lambda t: get_namespaced_index(request, t), result)
            indexes.extend(namespaced_results)
    # remove any duplicates
    indexes = list(set(indexes))
    index_string = ','.join(indexes)
    return index_string


def make_search_subreq(request, path):
    subreq = make_subrequest(request, path)
    if hasattr(request, "_stats"):
        subreq._stats = request._stats
    subreq.registry = request.registry
    if hasattr(request, "context"):
        subreq.context = request.context
    else:
        subreq.context = None
    subreq.headers['Accept'] = 'application/json'
    return subreq


DEFAULT_BROWSE_PARAM_LISTS = {
    'type': ["ExperimentSetReplicate"],
    'experimentset_type': ['replicate'],
    'award.project': ['4DN']
}


def get_iterable_search_results(request, search_path='/search/', param_lists=None, **kwargs):
    """
    Loops through search results, returns 100 (or search_results_chunk_row_size) results at a time.
    Pass it through itertools.chain.from_iterable to get one big iterable of results.
    TODO: Maybe make 'limit=all', and instead of calling invoke_subrequest(subrequest), instead call iter_search_results!

    :param request: Only needed to pass to do_subreq to make a subrequest with.
    :param search_path: Root path to call, defaults to /search/ (can also use /browse/).
    :param param_lists: Dictionary of param:lists_of_vals which is converted to URL query.
    :param search_results_chunk_row_size: Amount of results to get per chunk. Default should be fine.
    """
    if param_lists is None:
        param_lists = deepcopy(DEFAULT_BROWSE_PARAM_LISTS)
    else:
        param_lists = deepcopy(param_lists)
    param_lists['limit'] = ['all']
    param_lists['from'] = [0]  # TODO: Should be ['0'] ??
    param_lists['sort'] = param_lists.get('sort', 'uuid')  # TODO: Default should be ['uuid'] ??
    subreq = make_search_subreq(request, f'{search_path}?{urlencode(param_lists, True)}')
    return iter_search_results(None, subreq, **kwargs)


# Update? used in ./batch_download.py
def iter_search_results(context, request, **kwargs):
    return search(context, request, return_generator=True, **kwargs)


def build_table_columns(request, schemas, doc_types):

    any_abstract_types = 'Item' in doc_types
    if not any_abstract_types:  # Check explictly-defined types to see if any are abstract.
        type_infos = [request.registry[TYPES][type] for type in doc_types if type != 'Item']
        for ti in type_infos:
            # We use `type` instead of `isinstance` since we don't want to catch subclasses.
            if type(ti) == AbstractTypeInfo:
                any_abstract_types = True
                break

    columns = OrderedDict()

    # Add title column, at beginning always
    columns['display_title'] = {
        "title": "Title",
        "order": -100
    }

    # Add type column if any abstract types in search
    if any_abstract_types and request.normalized_params.get('currentAction') != 'selection':
        columns['@type'] = {
            "title": "Item Type",
            "colTitle": "Type",
            "order": -80,
            "description": "Type or category of Item",
            # Alternative below, if we want type column to be available but hidden by default in selection mode:
            # "default_hidden": request.normalized_params.get('currentAction') == 'selection'
        }

    for schema in schemas:
        if 'columns' in schema:
            schema_columns = OrderedDict(schema['columns'])
            # Add all columns defined in schema
            for name, obj in schema_columns.items():
                if name not in columns:
                    columns[name] = obj
                else:
                    # If @type or display_title etc. column defined in schema, then override defaults.
                    for prop in schema_columns[name]:
                        columns[name][prop] = schema_columns[name][prop]
                # Add description from field schema, if none otherwise.
                if not columns[name].get('description'):
                    field_schema = schema_for_field(name, request, doc_types)
                    if field_schema:
                        if field_schema.get('description') is not None:
                            columns[name]['description'] = field_schema['description']

    # Add status column, if not present, at end.
    if 'status' not in columns:
        columns['status'] = {
            "title": "Status",
            "default_hidden": True,
            "order": 501
        }
    # Add date column, if not present, at end.
    if 'date_created' not in columns:
        columns['date_created'] = {
            "title": "Date Created",
            "colTitle": "Created",
            "default_hidden": True,
            "order": 510
        }
    return columns


_ASSEMBLY_MAPPER = {
    'GRCh38-minimal': 'hg38',
    'GRCh38': 'hg38',
    'GRCh37': 'hg19',
    'GRCm38': 'mm10',
    'GRCm37': 'mm9',
    'BDGP6': 'dm4',
    'BDGP5': 'dm3',
    'WBcel235': 'WBcel235'
}

hgConnect = ''.join([
    'http://genome.ucsc.edu/cgi-bin/hgTracks',
    '?hubClear=',
])



# See http://docs.pylonsproject.org/projects/pyramid/en/latest/narr/resources.html
import logging
from collections import Mapping
from copy import deepcopy
from dcicutils.misc_utils import ignored
from pyramid.decorator import reify
from pyramid.httpexceptions import HTTPInternalServerError
from pyramid.security import (
    # Allow,
    Everyone,
    Authenticated,
    principals_allowed_by_permission
)
from pyramid.traversal import (
    resource_path,
    traverse
)
from .calculated import (
    calculate_properties,
    calculated_property,
)
from .interfaces import (
    COLLECTIONS,
    CONNECTION,
    ROOT,
    TYPES,
    UPGRADER,
)
from .validation import ValidationFailure
from .util import (
    ensurelist,
    simple_path_ids,
    uuid_to_path
)
from .util import add_default_embeds, IndexSettings

logger = logging.getLogger(__name__)


def includeme(config):
    config.include(auth0_config)
    config.scan(__name__)


class Resource(object):
    """
    Just used to add global calculated properties
    """
    @calculated_property(name='@id', schema={
        "title": "ID",
        "type": "string",
    })
    def jsonld_id(self, request):
        return request.resource_path(self)

    @calculated_property(name='@context', category='page')
    def jsonld_context(self, request):
        return request.route_path('jsonld_context')

    @calculated_property(category='page')
    def actions(self, request):
        actions = calculate_properties(self, request, category='action')
        if actions:
            return list(actions.values())


class Root(Resource):
    __name__ = ''
    __parent__ = None
    properties = {}

    def __init__(self, registry):
        self.registry = registry

    @reify
    def connection(self):
        return self.registry[CONNECTION]

    @reify
    def collections(self):
        return self.registry[COLLECTIONS]

    def __getitem__(self, name):
        try:
            resource = self.get(name)
        except KeyError:
            # Just in case we get an unexpected KeyError
            # FIXME: exception logging.
            raise HTTPInternalServerError('Traversal raised KeyError')
        if resource is None:
            raise KeyError(name)
        return resource

    def __contains__(self, name):
        return self.get(name, None) is not None

    def get(self, name, default=None):
        """
        Underlying get function used in traversal. Handles Collections (by
        direct `get`) and Items (through `Connection.__getitem__`)
        """
        resource = self.collections.get(name)
        if resource is not None:
            return resource
        try:
            resource = self.connection[name]  # Connection.__getitem__
        except KeyError:
            resource = default
        return resource

    def __json__(self, request=None):
        ignored(request)
        return self.properties.copy()

    @calculated_property(name='@type', schema={
        "title": "Type",
        "type": "array",
        "items": {
            "type": "string",
        },
    })
    def jsonld_type(self):
        return ['Portal']


def auth0_config(config):
    """ Route that exposes the auth0 domain and client ID for this app. """
    config.add_route(
        'auth0-config',
        '/auth0_config'
    )
    auth0_config_values = {  # determines which values are echoed
        'auth0.client': 'auth0Client',
        'auth0.domain': 'auth0Domain',
        'auth0.options': 'auth0Options'
    }

    def auth0_config_view(request):
        response = request.response
        response.content_type = 'application/json; charset=utf-8'
        response_dict = {
            'title': 'Auth0 Config',
        }
        settings = config.registry.settings
        for config_key, result_key in auth0_config_values.items():
            if config_key in settings:
                response_dict[result_key] = settings[config_key]
        return response_dict

    config.add_view(auth0_config_view, route_name='auth0-config')


class AbstractCollection(Resource, Mapping):
    """
    Collection for a certain type of resource that stores the following info:
    - registry (pyramid registry)
    - type_info (TypeInfo for a certain item type, see snovault.typeinfo.py)
    - __acl__
    - uniqueKey for the collection (e.g. item_name:key)
    - properties_datastore for items in the collection (defaults to 'database')
    And some other info as well.

    Collections allow retrieval of specific items with them by using the `get`
    method with uuid or the unique_key
    """
    properties = {}
    unique_key = None
    default_properties_datastore = 'database'

    def __init__(self, registry, name, type_info, properties=None, acl=None,
                 unique_key=None, properties_datastore=None):
        self.registry = registry
        self.__name__ = name
        self.type_info = type_info
        if properties is not None:
            self.properties = properties
        if acl is not None:
            self.__acl__ = acl
        if unique_key is not None:
            self.unique_key = unique_key
        # use a default value configured on the class if not set
        if properties_datastore is not None:
            self.properties_datastore = properties_datastore
        else:
            self.properties_datastore = self.default_properties_datastore

    @reify
    def connection(self):
        return self.registry[CONNECTION]

    @reify
    def __parent__(self):
        return self.registry[ROOT]

    def __getitem__(self, name):
        try:
            item = self.get(name)
        except KeyError:
            # Just in case we get an unexpected KeyError
            # FIXME: exception logging.
            raise HTTPInternalServerError('Traversal raised KeyError')
        if item is None:
            raise KeyError(name)
        return item

    def __iter__(self):
        for uuid in self.connection.__iter__(*self.type_info.subtypes):
            yield uuid

    def __len__(self):
        return self.connection.__len__(*self.type_info.subtypes)

    def __hash__(self):
        return object.__hash__(self)

    def __eq__(self, other):
        return self is other

    def _allow_contained(self, resource):
        return resource.__parent__ is self or \
            resource.type_info.name in resource.type_info.subtypes

    def get(self, name, default=None):
        """
        Get an item by name using this collection. First try uuid and then
        unique key. If neither are found, return default, which is usually None.
        self.properties_datastore is checked on the item class and used with
        Connection to force a certain datastore if needed to retrieve item
        """
        resource = self.connection.get_by_uuid(name, default=None,
                                               datastore=self.properties_datastore)
        if resource is not None:
            if not self._allow_contained(resource):
                return default
            return resource
        if self.unique_key is not None:
            item_type_snake_case = getattr(self.type_info, 'item_type', None)
            resource = self.connection.get_by_unique_key(self.unique_key, name,
                                                         datastore=self.properties_datastore,
                                                         item_type=item_type_snake_case)
            if resource is not None:
                if not self._allow_contained(resource):
                    return default
                return resource
        return default

    def iter_no_subtypes(self):
        """
        Make a generator that yields all items in the collection, but not
        subtypes
        """
        for uuid in self.connection.__iter__(self.type_info.item_type):
            yield uuid

    def __json__(self, request):
        ignored(request)
        return self.properties.copy()

    @calculated_property(name='@type', schema={
        "title": "Type",
        "type": "array",
        "items": {
            "type": "string",
        },
    })
    def jsonld_type(self):
        return [
            '{type_name}Collection'.format(type_name=self.type_info.name),
            'Collection',
        ]


class Collection(AbstractCollection):
    """ Separate class so add views do not apply to AbstractCollection """

    @staticmethod
    def index_settings():
        """ Corresponds to the default index settings prior to the existence of this property """
        return IndexSettings()


# Almost every single display_title should have the same
# schema definition, so we define it here to import & re-use.
display_title_schema = {
    "title": "Display Title",
    "description": "A calculated title for every object",
    "type": "string",
}


class Item(Resource):
    """
    Base Item resource that corresponds to a Collection or AbstractCollection
    """
    item_type = None
    base_types = ['Item']
    name_key = None
    rev = {}
    aggregated_items = {}
    embedded_list = []
    default_diff = []
    filtered_rev_statuses = ()
    schema = None
    AbstractCollection = AbstractCollection
    Collection = Collection

    def __init__(self, registry, model):
        self.registry = registry
        self.model = model

    def __repr__(self):
        return '<%s at %s>' % (type(self).__name__, resource_path(self))

    @reify
    def type_info(self):
        return self.registry[TYPES][type(self)]

    @reify
    def collection(self):
        collections = self.registry[COLLECTIONS]
        return collections[self.type_info.name]

    @reify
    def properties_datastore(self):
        return self.collection.properties_datastore

    @reify
    def default_properties_datastore(self):
        return self.collection.default_properties_datastore

    @property
    def __parent__(self):
        return self.collection

    @property
    def __name__(self):
        """
        Used in the resource path for this item. Use `self.name_key` if
        present, otherwise `self.uuid`
        """
        if self.name_key is None:
            return str(self.uuid)
        return self.properties.get(self.name_key, None) or str(self.uuid)

    @property
    def properties(self):
        return self.model.properties

    @property
    def propsheets(self):
        return self.model.propsheets

    @property
    def uuid(self):
        return self.model.uuid

    @property
    def sid(self):
        return self.db_model.sid

    @property
    def max_sid(self):
        return self.db_model.max_sid

    @property
    def db_model(self):
        """
        Always returns the resouce model from write storage, which is needed
        for operations like getting current sid/max_sid, rev_links, and
        updating. Leverage `model.used_datastore` to determine source
        """
        if self.model.used_datastore != 'database':
            connection = self.registry[CONNECTION]
            return connection.storage.write.get_by_uuid(str(self.uuid))
        return self.model

    def links(self, properties):
        return {
            path: list(set(simple_path_ids(properties, path)))
            for path in self.type_info.schema_links
        }

    def get_rev_links(self, request, name):
        """
        Return a list of uuid rev_links for the given rev name (in self.rev)
        from the given item.

        Args:
            request: current Request
            name (str): name of the rev (must be in self.rev)

        Returns:
            list of str uuids of the given rev_link
        """
        ignored(request)
        types = self.registry[TYPES]
        type_name, rel = self.rev[name]
        types = types[type_name].subtypes
        return self.registry[CONNECTION].get_rev_links(self.db_model, rel, *types)

    def get_filtered_rev_links(self, request, name):
        """
        Run get_rev_links, but only return items that do not have a status
        in self.filtered_rev_statuses (a tuple defined on the Item)
        If we are indexing, add rev_link info to _rev_linked_uuids_by_item.

        Args:
            request: current Request
            name (str): name of the rev (must be in self.rev)

        Returns:
            list of str uuids of the given rev_link, filtered by status
        """
        # Consider caching rev links on the request? Would save DB requests
        # May not be worth it because they are quite fast
        rev_uuids = self.get_rev_links(request, name)
        filtered_uuids = [
            str(rev_id) for rev_id in rev_uuids
            if traverse(request.root, str(rev_id))['context'].__json__(request).get('status')
            not in self.filtered_rev_statuses
        ]
        if request._indexing_view is True:
            to_update = {name: filtered_uuids}
            if str(self.uuid) in request._rev_linked_uuids_by_item:
                request._rev_linked_uuids_by_item[str(self.uuid)].update(to_update)
            else:
                request._rev_linked_uuids_by_item[str(self.uuid)] = to_update
        return filtered_uuids

    def unique_keys(self, properties):
        return {
            name: [v for prop in props for v in ensurelist(properties.get(prop, ()))]
            for name, props in self.type_info.schema_keys.items()
        }

    def upgrade_properties(self):
        """
        Calls the upgrader on the Item if properties.schema_version is not current
        """
        try:
            properties = deepcopy(self.properties)
        except KeyError:
            # don't fail if we try to upgrade properties on something not there yet
            return None
        current_version = properties.get('schema_version', '1')
        target_version = self.type_info.schema_version
        if target_version is not None and current_version != target_version:
            upgrader = self.registry[UPGRADER]
            try:
                properties = upgrader.upgrade(
                    self.type_info.name, properties, current_version, target_version,
                    context=self, registry=self.registry)
            except RuntimeError:
                raise
            except Exception:
                logger.warning(
                    'Unable to upgrade %s from %r to %r',
                    resource_path(self.__parent__, self.uuid),
                    current_version, target_version, exc_info=True)
        return properties

    def __json__(self, request):
        """
        This function is used to get the "complete" properties of the Item
        after calling `upgrade_properties`
        """
        ignored(request)
        return self.upgrade_properties()

    def item_with_links(self, request):
        """
        Key function that transforms uuids into resource paths in properties.
        It is also responsible for calling upgraders using the __json__ method

        VERY IMPORTANT NOTE:
            Lastly, adds this items' uuid to request._linked_uuids when indexing
            This is now a tuple of uuid, item_type so when retrieving links from ES
            for all links we know both the uuid AND the type.
        """
        # *** context.__json__ CALLS THE UPGRADER (upgrade_properties) ***
        # This works from the schema rather than the links table
        # so that upgrade on GET can work.
        properties = self.__json__(request)
        # use schema_links rather than DB links so upgrades work on ES GETs
        for path in self.type_info.schema_links:
            uuid_to_path(request, properties, path)

        # if indexing, add the uuid of this object to request._linked_uuids
        # and add the sid to _sid_cache if not already present
        if request._indexing_view is True:
            request._linked_uuids.add((str(self.uuid), self.type_info.name))  # _linked_uuids holds 2 tuple
            if str(self.uuid) not in request._sid_cache:
                request._sid_cache[str(self.uuid)] = self.sid
        return properties

    def __resource_url__(self, request, info):
        ignored(request, info)
        return None

    @classmethod
    def create(cls, registry, uuid, properties, sheets=None):
        """
        This class method is called in crud_views.py - `collection_add` (API endpoint) > `create_item` (method) > `type_info.factory.create` (this class method)

        This method instantiates a new Item class instance from provided `uuid` and `properties`,
        then runs the `_update` (instance method) to save the Item to the database.
        """
        model = registry[CONNECTION].create(cls.__name__, uuid)
        item_instance = cls(registry, model)
        item_instance._update(properties, sheets)
        return item_instance

    def validate_path_characters(self, field, value):
        """
        Check that the field with given value does not contain any characters
        that interfere with the resource_path. Currently, we allow all
        alphanumeric characters and few others
        """
        also_allowed = ['_', '-', ':', ',', '.', ' ', '@']
        if not isinstance(value, str):  # formerly basestring
            raise ValueError('Identifying property %s must be a string. Value: %s' % (field, value))
        forbidden = [char for char in value
                     if (not char.isalnum() and char not in also_allowed)]
        if any(forbidden):
            msg = ("Forbidden character(s) %s are not allowed in field: %s. Value: %s"
                   % (set(forbidden), field, value))
            raise ValidationFailure('body', 'Item: path characters', msg)

    def update(self, properties, sheets=None):
        """Alias of _update, called in crud_views.py - `update_item` (method)"""
        self._update(properties, sheets)

    def _update(self, properties, sheets=None):
        """
        This instance method is called in Item.create (classmethod) as well as in crud_views.py - `item_edit` (API endpoint) > `update_item` (method) > `context.update` (instance method).

        This method is used to assert lack of duplicate unique keys in database and then to perform database update of `properties` (dict).

        Optionally define this method in inherited classes to extend `properties` on Item updates.
        """
        unique_keys = None
        links = None
        if properties is not None:
            if 'uuid' in properties:
                properties = properties.copy()
                del properties['uuid']

            # validation on name key and unique keys
            nk_val = properties.get(self.name_key, '')
            self.validate_path_characters(self.name_key, nk_val)

            unique_keys = self.unique_keys(properties)
            for k, values in unique_keys.items():
                if len(set(values)) != len(values):
                    msg = "Duplicate keys for %r: %r" % (k, values)
                    raise ValidationFailure('body', 'Item: duplicate keys', msg)
                for uk_val in values:
                    self.validate_path_characters(k, uk_val)

            links = self.links(properties)

        # actually propogate the update to the DB
        connection = self.registry[CONNECTION]
        connection.update(self.db_model, properties, sheets, unique_keys, links,
                          datastore=self.properties_datastore)

    @reify
    def embedded(self):
        """
        Use the embedded_list defined for the individual types to create the
        embedded attribute through expansion using add_default_embeds
        """
        total_schema = self.schema.get('properties', {}).copy()
        calc_props_schema = {}
        types = self.registry[TYPES]
        if self.registry['calculated_properties']:
            for calc_props_key, calc_props_val in self.registry['calculated_properties'].props_for(self).items():
                if calc_props_val.schema:
                    calc_props_schema[calc_props_key] = calc_props_val.schema
        total_schema.update(calc_props_schema)
        this_type = self.type_info.item_type
        return add_default_embeds(this_type, types, self.embedded_list, total_schema)

    @calculated_property(name='@type', schema={
        "title": "Type",
        "type": "array",
        "items": {
            "type": "string",
        },
    })
    def jsonld_type(self):
        return [self.type_info.name] + self.base_types

    @calculated_property(name='uuid')
    def prop_uuid(self):
        return str(self.uuid)

    @calculated_property(schema={
        "title": "principals_allowed",
        "description": "Calculated permissions used for ES filtering",
        "type": "object",
        'properties': {
            'view': {
                'type': 'string'
            },
            'edit': {
                'type': 'string'
            }
        }
    })
    def principals_allowed(self):
        allowed = {}
        # these are the relevant Item permissions
        for permission in ('view', 'edit'):
            principals = principals_allowed_by_permission(self, permission)
            if principals is Everyone:
                principals = [Everyone]
            elif Everyone in principals:
                principals = [Everyone]
            elif Authenticated in principals:
                principals = [Authenticated]
            # Filter our roles
            allowed[permission] = [
                p for p in sorted(principals) if not p.startswith('role.')
            ]
        return allowed

    @calculated_property(schema=display_title_schema)
    def display_title(self):
        return str(self.uuid)



"""\
Examples

To check all pages on the production server:

    %(prog)s production.ini

For the development.ini you must supply the paster app name:

    %(prog)s development.ini --app-name app

"""

import argparse
import json
import logging
import webtest

from future.utils import itervalues
from pyramid import paster
from pyramid.traversal import resource_path


EPILOG = __doc__

logger = logging.getLogger(__name__)

def check_path(testapp, path):
    try:
        res = testapp.get(path, status='*').maybe_follow(status='*')
    except Exception:
        logger.exception('Render failed: %s', path)
        return False
    if res.status_int != 200:
        logger.error('Render failed (%s): %s', res.status, path)
        script = res.html.find('script', **{'data-prop-name': 'context'})
        if script is not None:
            context = json.loads(script.text)
            if 'detail' in context:
                logger.debug(context['detail'])
            else:
                logger.debug(json.dumps(context, indent=4))
        return False
    return True


def run(testapp, collections=None):
    app = testapp.app
    root = app.root_factory(app)
    if not collections:
        collections = root.by_item_type.keys()
        check_path(testapp, '/')
    for collection_name in collections:
        collection = root[collection_name]
        collection_path = resource_path(collection, '')
        check_path(testapp, collection_path)
        failed = 0
        for count, item in enumerate(itervalues(collection)):
            path = resource_path(item, '')
            if not check_path(testapp, path):
                failed += 1
        if failed:
            logger.info('Collection %s: %d of %d failed to render.',
                collection_path, failed, count)
        else:
            logger.info('Collection %s: all %d rendered ok',
                collection_path, count)


def internal_app(configfile, app_name=None, username='TEST', accept='text/html'):
    app = paster.get_app(configfile, app_name)
    environ = {
        'HTTP_ACCEPT': accept,
        'REMOTE_USER': username,
    }
    return webtest.TestApp(app, environ)


def main():
    parser = argparse.ArgumentParser(
        description="Check rendering of items", epilog=EPILOG,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument('--item-type', action='append', help="Item type")
    parser.add_argument('--app-name', help="Pyramid app name in configfile")
    parser.add_argument('--username', '-u', default='TEST',
        help="User uuid/email")
    parser.add_argument('config_uri', help="path to configfile")
    parser.add_argument('path', nargs='*', help="path to test")
    args = parser.parse_args()

    logging.basicConfig()
    testapp = internal_app(args.config_uri, args.app_name, args.username)
    # Loading app will have configured from config file. Reconfigure here:
    logging.getLogger('snovault').setLevel(logging.DEBUG)

    if args.path:
        failed = 0
        for path in args.path:
            if not check_path(testapp, path):
                failed += 1
        if failed:
            logger.info('Paths: %d of %d failed to render.',
                failed, len(args.path))
        else:
            logger.info('Paths: all %d rendered ok', len(args.path))
    else:
        run(testapp, args.item_type)


if __name__ == '__main__':
    main()



import base64
import os

from dcicutils.misc_utils import ignored
from operator import itemgetter
from passlib.context import CryptContext
from pyramid.authentication import (
    BasicAuthAuthenticationPolicy as _BasicAuthAuthenticationPolicy,
    CallbackAuthenticationPolicy
)
from pyramid.httpexceptions import HTTPForbidden, HTTPFound
from pyramid.path import caller_package, DottedNameResolver
from pyramid.security import (
    # Authenticated,
    # Everyone,
    forget,
    NO_PERMISSION_REQUIRED,
    remember,
    # principals_allowed_by_permission,
)
from pyramid.settings import asbool  # , aslist
from pyramid.view import view_config
from ..interfaces import ROOT, COLLECTIONS
from ..storage import User
from ..calculated import calculate_properties
from ..validation import ValidationFailure
from ..validators import no_validate_item_content_post


CRYPT_CONTEXT = __name__ + ':crypt_context'


def includeme(config):
    config.include('.snowflake_hash')
    setting_prefix = 'passlib.'
    passlib_settings = {
        k[len(setting_prefix):]: v
        for k, v in config.registry.settings.items()
        if k.startswith(setting_prefix)
    }
    if not passlib_settings:
        passlib_settings = {'schemes': 'snowflake_hash, unix_disabled'}
    crypt_context = CryptContext(**passlib_settings)
    config.registry[CRYPT_CONTEXT] = crypt_context

    # basic login route
    config.add_route('login', '/login')
    config.add_route('logout', '/logout')
    config.add_route('impersonate-user', '/impersonate-user')
    config.add_route('session-properties', '/session-properties')
    config.scan(__name__)


class NamespacedAuthenticationPolicy(object):
    """ Wrapper for authentication policy classes

    As userids are included in the list of principals, it seems good practice
    to namespace them to avoid clashes.

    Constructor Arguments

    ``namespace``

        The namespace used (string).

    ``base``

        The base authentication policy (class or dotted name).

    Remaining arguments are passed to the ``base`` constructor.

    Example

    To make a ``REMOTE_USER`` 'admin' be 'user.admin'

    .. code-block:: python

        policy = NamespacedAuthenticationPolicy('user',
            'pyramid.authentication.RemoteUserAuthenticationPolicy')
    """

    def __new__(cls, namespace, base, *args, **kw):
        # Dotted name support makes it easy to configure with pyramid_multiauth
        name_resolver = DottedNameResolver(caller_package())
        base = name_resolver.maybe_resolve(base)
        # Dynamically create a subclass
        name = 'Namespaced_%s_%s' % (namespace, base.__name__)
        klass = type(name, (cls, base), {'_namespace_prefix': namespace + '.'})
        return super(NamespacedAuthenticationPolicy, klass).__new__(klass)

    def __init__(self, namespace, base, *args, **kw):
        ignored(namespace, base)
        super(NamespacedAuthenticationPolicy, self).__init__(*args, **kw)

    def unauthenticated_userid(self, request):
        userid = super(NamespacedAuthenticationPolicy, self) \
            .unauthenticated_userid(request)
        if userid is not None:
            userid = self._namespace_prefix + userid
        return userid

    def remember(self, request, principal, **kw):
        if not principal.startswith(self._namespace_prefix):
            return []
        principal = principal[len(self._namespace_prefix):]
        return super(NamespacedAuthenticationPolicy, self) \
            .remember(request, principal, **kw)


class BasicAuthAuthenticationPolicy(_BasicAuthAuthenticationPolicy):
    def __init__(self, check, *args, **kw):
        # Dotted name support makes it easy to configure with pyramid_multiauth
        name_resolver = DottedNameResolver(caller_package())
        check = name_resolver.maybe_resolve(check)
        super(BasicAuthAuthenticationPolicy, self).__init__(check, *args, **kw)


class LoginDenied(HTTPForbidden):
    title = 'Login failure'


_fake_user = object()


class WebUserAuthenticationPolicy(CallbackAuthenticationPolicy):

    login_path = '/login'
    method = 'POST'

    def unauthenticated_userid(self, request):
        """
        So basically this is used to do a login, instead of the actual
        login view... not sure why, but yeah..
        """
        # if we aren't posting to login just return None
        if request.method != self.method or request.path != self.login_path:
            return None

        # otherwise do a login, if we aren't already logged in
        cached = getattr(request, '_webuser_authenticated', _fake_user)
        if cached is not _fake_user:
            return cached

        login = request.json.get("username")
        password = request.json.get("password")
        if not User.check_password(login, password):
            request._webuser_authenticated = None
            return None

        request._webuser_authenticated = login
        return login

    @classmethod
    def remember(cls, request, principal, **kw):
        ignored(request, principal, kw)
        return []

    @classmethod
    def forget(cls, request):
        ignored(request)
        return []


@view_config(route_name='login', request_method='POST',
             permission=NO_PERMISSION_REQUIRED)
def login(request):
    login = request.authenticated_userid
    if login is None:
        namespace = userid = None
    else:
        namespace, userid = login.split('.', 1)

    if namespace != 'webuser':
        request.session.invalidate()
        request.response.headerlist.extend(forget(request))
        raise LoginDenied()

    request.session.invalidate()
    request.session.get_csrf_token()
    request.response.headerlist.extend(remember(request, 'mailto.' + userid))

    properties = request.embed('/session-properties', as_user=userid)
    if 'auth.userid' in request.session:
        properties['auth.userid'] = request.session['auth.userid']
    return properties


@view_config(route_name='logout',
             permission=NO_PERMISSION_REQUIRED, http_cache=0)
def logout(request):
    """View to forget the user"""
    request.session.invalidate()
    request.session.get_csrf_token()
    request.response.headerlist.extend(forget(request))
    if asbool(request.params.get('redirect', True)):
        raise HTTPFound(location=request.resource_path(request.root))
    return {}


@view_config(route_name='session-properties', request_method='GET',
             permission=NO_PERMISSION_REQUIRED)
def session_properties(request):
    for principal in request.effective_principals:
        if principal.startswith('userid.'):
            break
    else:
        return {}

    namespace, userid = principal.split('.', 1)
    user = request.registry[COLLECTIONS]['user'][userid]
    user_actions = calculate_properties(user, request, category='user_action')

    properties = {
        'user': request.embed(request.resource_path(user)),
        'user_actions': [v for k, v in sorted(user_actions.items(), key=itemgetter(0))]
    }

    if 'auth.userid' in request.session:
        properties['auth.userid'] = request.session['auth.userid']

    return properties


def webuser_check(username, password, request):
    ignored(request)
    # webusers have email address for username, thus, make sure we have an email address
    if '@' not in username:
        return None
    if not User.check_password(username, password):
        return None
    return []


def basic_auth_check(username, password, request):
    # We may get called before the context is found and the root set
    root = request.registry[ROOT]
    collection = root['access-keys']
    try:
        access_key = collection[username]
    except KeyError:
        return None

    properties = access_key.properties
    hash = properties['secret_access_key_hash']

    crypt_context = request.registry[CRYPT_CONTEXT]
    valid = crypt_context.verify(password, hash)
    if not valid:
        return None

    # valid, new_hash = crypt_context.verify_and_update(password, hash)
    # if new_hash:
    #     replace_user_hash(user, new_hash)

    return []


@view_config(route_name='impersonate-user', request_method='POST',
             validators=[no_validate_item_content_post],
             permission='impersonate')
def impersonate_user(request):
    """As an admin, impersonate a different user."""
    userid = request.validated['userid']
    users = request.registry[COLLECTIONS]['user']

    try:
        user = users[userid]
    except KeyError:
        raise ValidationFailure('body', ['userid'], 'User not found.')

    if user.properties.get('status') != 'current':
        raise ValidationFailure('body', ['userid'], 'User is not enabled.')

    request.session.invalidate()
    request.session.get_csrf_token()
    request.response.headerlist.extend(remember(request, 'mailto.' + userid))
    user_properties = request.embed('/session-properties', as_user=userid)
    if 'auth.userid' in request.session:
        user_properties['auth.userid'] = request.session['auth.userid']

    return user_properties


def generate_user():
    """ Generate a random user name with 64 bits of entropy
        used to generate access_key
        remove @ to ensure differentiation from web users, see webuser_check
    """
    # Take a random 5 char binary string (80 bits of
    # entropy) and encode it as upper cased base32 (8 chars)
    random_bytes = os.urandom(5)
    user = base64.b32encode(random_bytes).decode('ascii').rstrip('=').upper()
    user = user.replace("@", "")
    return user


def generate_password():
    """ Generate a password with 80 bits of entropy
    """
    # Take a random 10 char binary string (80 bits of
    # entropy) and encode it as lower cased base32 (16 chars)
    random_bytes = os.urandom(10)
    password = base64.b32encode(random_bytes).decode('ascii').rstrip('=').lower()
    return password



import logging
import pytest
import sqlalchemy
import sqlalchemy.orm
import transaction as transaction_management
import webtest
import webtest.http
import zope.sqlalchemy

from contextlib import contextmanager
from dcicutils.misc_utils import ignored, environ_bool
from dcicutils.qa_utils import notice_pytest_fixtures
from transaction.interfaces import ISynchronizer
from zope.interface import implementer

from ..app import configure_engine
from ..storage import Base
from .elasticsearch_fixture import server_process as elasticsearch_server_process
from .postgresql_fixture import (
    initdb, server_process as postgres_server_process,
    # SNOVAULT_DB_TEST_PORT,
    make_snovault_db_test_url,
)


NO_SERVER_FIXTURES = environ_bool("NO_SERVER_FIXTURES")

def pytest_configure():
    logging.basicConfig(format='')
    logging.getLogger('sqlalchemy.engine').setLevel(logging.WARNING)

    class Shorten(logging.Filter):
        max_len = 500

        def filter(self, record):
            if record.msg == '%r':
                record.msg = record.msg % record.args
                record.args = ()
            if len(record.msg) > self.max_len:
                record.msg = record.msg[:self.max_len] + '...'
            return True

    logging.getLogger('sqlalchemy.engine.base.Engine').addFilter(Shorten())


@pytest.mark.fixture_cost(10)
@pytest.yield_fixture(scope='session')
def engine_url(tmpdir_factory):
    if NO_SERVER_FIXTURES:
        yield 'NO_SERVER_FIXTURES'
        return

    notice_pytest_fixtures(tmpdir_factory)

    # Ideally this would use a different database on the same postgres server
    tmpdir = tmpdir_factory.mktemp('postgresql-engine', numbered=True)
    tmpdir = str(tmpdir)
    initdb(tmpdir)
    process = postgres_server_process(tmpdir)

    yield make_snovault_db_test_url(datadir=tmpdir)

    if process.poll() is None:
        process.terminate()
        process.wait()


@pytest.mark.fixture_cost(10)
@pytest.yield_fixture(scope='session')
def postgresql_server(tmpdir_factory):
    if NO_SERVER_FIXTURES:
        yield 'NO_SERVER_FIXTURES'
        return

    notice_pytest_fixtures(tmpdir_factory)
    tmpdir = tmpdir_factory.mktemp('postgresql', numbered=True)
    tmpdir = str(tmpdir)
    initdb(tmpdir)
    process = postgres_server_process(tmpdir)

    yield make_snovault_db_test_url(datadir=tmpdir)

    if process.poll() is None:
        process.terminate()
        # Should there be a process.wait() here? -kmp 14-Mar-2020


@pytest.fixture(scope='session')
def elasticsearch_host_port():
    return webtest.http.get_free_port()


@pytest.fixture(scope='session')
def elasticsearch_server_dir(tmpdir_factory):
    return str(tmpdir_factory.mktemp('elasticsearch', numbered=True))


@pytest.mark.fixture_cost(10)
@pytest.yield_fixture(scope='session')
def elasticsearch_server(elasticsearch_server_dir, elasticsearch_host_port, remote_es):
    if NO_SERVER_FIXTURES:
        yield 'NO_SERVER_FIXTURES'
        return

    notice_pytest_fixtures(elasticsearch_host_port, remote_es)
    if not remote_es:
        # spawn a new one
        host, port = elasticsearch_host_port
        process = elasticsearch_server_process(elasticsearch_server_dir, host=host, port=port)

        yield 'http://%s:%d' % (host, port)

        if process.poll() is None:
            process.terminate()
            process.wait()
    else:
        yield remote_es


# http://docs.sqlalchemy.org/en/rel_0_8/orm/session.html#joining-a-session-into-an-external-transaction
# By binding the SQLAlchemy Session to an external transaction multiple testapp
# requests can be rolled back at the end of the test.

@pytest.yield_fixture(scope='session')
def engine(engine_url):
    if NO_SERVER_FIXTURES:
        yield 'NO_SERVER_FIXTURES'
        return

    notice_pytest_fixtures(engine_url)
    engine_settings = {
        'sqlalchemy.url': engine_url,
        'connect_args': {
            'connection_timeout': 30
        }
    }

    engine = configure_engine(engine_settings)
    yield engine


@pytest.yield_fixture(scope='session')
def conn(engine):
    if NO_SERVER_FIXTURES:
        yield 'NO_SERVER_FIXTURES'
        return

    notice_pytest_fixtures(engine)
    conn = engine.connect()
    tx = conn.begin()
    try:
        Base.metadata.create_all(bind=conn)
        yield conn
    finally:
        tx.rollback()
        conn.close()
        engine.dispose()


@pytest.fixture(scope='session')
def _DBSession(conn):
    if NO_SERVER_FIXTURES:
        return 'NO_SERVER_FIXTURES'

    notice_pytest_fixtures(conn)
    # ``server`` thread must be in same scope
    DBSession = sqlalchemy.orm.scoped_session(sqlalchemy.orm.sessionmaker(bind=conn), scopefunc=lambda: 0)
    zope.sqlalchemy.register(DBSession)
    return DBSession


@pytest.fixture(scope='session')
def DBSession(_DBSession, zsa_savepoints, check_constraints):
    if NO_SERVER_FIXTURES:
        return 'NO_SERVER_FIXTURES'

    notice_pytest_fixtures(zsa_savepoints, check_constraints)
    return _DBSession


@pytest.yield_fixture
def external_tx(request, conn):
    if NO_SERVER_FIXTURES:
        yield 'NO_SERVER_FIXTURES'
        return

    notice_pytest_fixtures(request)
    # print('BEGIN external_tx')
    tx = conn.begin_nested()
    try:
        yield tx
        tx.rollback()
    except Exception:
        tx.rollback()  # conn does not implement .rollback()
    # # The database should be empty unless a data fixture was loaded
    # for table in Base.metadata.sorted_tables:
    #     assert conn.execute(table.count()).scalar() == 0


@pytest.fixture
def transaction(request, external_tx, zsa_savepoints, check_constraints):
    if NO_SERVER_FIXTURES:
        return 'NO_SERVER_FIXTURES'

    notice_pytest_fixtures(request, external_tx, zsa_savepoints, check_constraints)
    transaction_management.begin()
    request.addfinalizer(transaction_management.abort)
    return transaction_management


@pytest.yield_fixture(scope='session')
def zsa_savepoints(conn):
    """ Place a savepoint at the start of the zope transaction

    This means failed requests rollback to the db state when they began rather
    than that at the start of the test.
    """
    notice_pytest_fixtures(conn)

    if NO_SERVER_FIXTURES:
        yield 'NO_SERVER_FIXTURES'
        return

    @implementer(ISynchronizer)
    class Savepoints(object):
        def __init__(self, conn):
            self.conn = conn
            self.sp = None
            self.state = None

        def beforeCompletion(self, transaction):
            pass

        def afterCompletion(self, transaction):
            # txn be aborted a second time in manager.begin()
            if self.sp is None:
                return
            if self.state == 'commit':
                self.state = 'completion'
                # self.sp.commit() this call is not necessary and throwing exception in sqlalchemy
            else:
                self.state = 'abort'
                self.sp.rollback()
            self.sp = None
            self.state = 'done'

        def newTransaction(self, transaction):
            self.state = 'new'
            self.sp = self.conn.begin_nested()
            self.state = 'begun'
            transaction.addBeforeCommitHook(self._registerCommit)

        def _registerCommit(self):
            self.state = 'commit'

    zsa_savepoints = Savepoints(conn)

    transaction_management.manager.registerSynch(zsa_savepoints)

    yield zsa_savepoints
    transaction_management.manager.unregisterSynch(zsa_savepoints)


@pytest.fixture
def session(transaction, DBSession):
    """ Returns a setup session

    Depends on transaction as storage relies on some interaction there.
    """
    if NO_SERVER_FIXTURES:
        return 'NO_SERVER_FIXTURES'

    notice_pytest_fixtures(transaction, DBSession)
    return DBSession()


@pytest.yield_fixture(scope='session')
def check_constraints(conn, _DBSession):
    """
    Check deferred constraints on zope transaction commit.

    Deferred foreign key constraints are only checked at the outer transaction
    boundary, not at a savepoint. With the Pyramid transaction bound to a
    subtransaction check them manually.
    """
    notice_pytest_fixtures(_DBSession)

    if NO_SERVER_FIXTURES:
        yield 'NO_SERVER_FIXTURES'
        return

    @implementer(ISynchronizer)
    class CheckConstraints(object):
        def __init__(self, conn):
            self.conn = conn
            self.state = None

        def beforeCompletion(self, transaction):
            pass

        def afterCompletion(self, transaction):
            pass

        def newTransaction(self, transaction):

            @transaction.addBeforeCommitHook
            def set_constraints():
                self.state = 'checking'
                session = _DBSession()
                session.flush()
                sp = self.conn.begin_nested()
                try:
                    self.conn.execute('SET CONSTRAINTS ALL IMMEDIATE')
                except BaseException:  # even things like keyboard interrupt
                    sp.rollback()
                    raise
                else:
                    self.conn.execute('SET CONSTRAINTS ALL DEFERRED')
                finally:
                    sp.commit()
                    self.state = None

    check_constraints = CheckConstraints(conn)

    transaction_management.manager.registerSynch(check_constraints)

    yield check_constraints

    transaction_management.manager.unregisterSynch(check_constraints)


class ExecutionWatcher(object):

    """
    An ExecutionWatcher mediates the counting of something, which can scoped by the .expect() context manager, as in:
            with <watcher>.expect(expect=n):
               ...watched stuff happens here...
    It will report an error if the .notify() method doesn't see exactly n events that later survive filtering.
    If the right number does not happen, the error message will say what events were seen in
    the (active) watched region.
    """

    def __init__(self, filter=None):
        """
        Creates an ExecutionWatcher object with a given filter.
        """

        # The ._active flag is used to know whether to be recording and also to make sure we don't do nested calls.
        # We are only recording if we are inside a 'with execute_counter.expect(...):' context, which is to say
        # this ExecutionWatcher is ordinarily created within a use of the `execute_counter` fixture. See documentation
        # on that fixture for details of use.

        self._active = False  # The wacher is not reentrant
        self.reset()
        self.filter = filter

    def reset(self):
        """
        This can be used to reinitialize counting, though the semantics of that are questionable.
        It MIGHT have limited utility in the case of a db rollback, but we really don't use that.
        """
        # TODO: It might be that this method could usefully go away.
        self.events = []

    def notice(self, event):
        """
        Whatever it is that's being instrumented should send events here to get them counted.
        Any filter will be done later, so that if events are being filtered that shouldn't be,
        they will still show up in the error message.
        """
        if self._active:
            self.events.append(event)

    @contextmanager
    def expect(self, expected_count):
        """
        This context manager declares that within a certain scope, only a certain number of events are expected.
        See documentation on the
        """
        if self._active:
            raise RuntimeError("Attempt to enter execute_counter.expect(...) while it is already executing.")
        self._active = True
        self.reset()  # Probably redundant but just in case.
        yield
        annotated_events = []
        counted_events = 0
        for event in self.events:
            to_count = (not self.filter) or self.filter(event)
            if to_count:
                counted_events += 1
            annotated_events.append({'counted': to_count, 'event': event})
        assert counted_events == expected_count, (
                f"Counter mismatch. Expected {expected_count} but got {counted_events}:\n"
                + '\n'.join([f"{'*' if ae['counted'] else ' '} {ae['event']}"
                            for ae in annotated_events]))
        self._active = False


@pytest.yield_fixture
def execute_counter(conn, zsa_savepoints, check_constraints, filter=None):
    """
    This fixture gives you a context manager that can be used to count calls to the SQLAlchemy 'execute' operation.

    Using this allows you to find out how many queries are needed to do a particular inquiry or set of inquiries
    using the ORM. Such a test is by nature a bit fragile, since it is bypassing abstraction boundaries, but I think
    the intent is to let the caller check whether joins are happening correctly. If they are not, the likely result
    is that you'll see more than the expected number of database queries.

    The intended use is in testing, where the fixture gives you a counter object that has a .expect method you can
    use as a context manager in order to bound the region of code you want to count. For example:

       def test_something(execute_counter):
           ... do some setup ...
           execute_counter.expect(2):  # <-- this is where you say how many DB executes you expect
               ... some ORM operation ...

    You'll get an error if the expectation is violated, and it will tell you what the operations were so that you can
    sort out whether you think it's an error or just a legit change in the underlying implementation.

    Note that counters are not reentrant. That is, you can't nest these calls with the same fixture. (If you call
    something else that has its own fixture, it should be fine.) There are no cases we care about where nesting was
    needed, so it simplifies the implementation.

    There is also now a filter argument to the execute_counter fixture that will allow, in the future if it is needed,
    for the possibility that some execute operations would be ignored (not counted) during the measured interval.
    (That was created for debugging and didn't end up being used, but still might be useful in the future.)
    """
    notice_pytest_fixtures(conn, zsa_savepoints, check_constraints)

    if NO_SERVER_FIXTURES:
        yield 'NO_SERVER_FIXTURES'
        return

    watcher = ExecutionWatcher(filter=filter)

    @sqlalchemy.event.listens_for(conn, 'after_cursor_execute')
    def after_cursor_execute(conn, cursor, statement, parameters, context, executemany):
        ignored(conn, cursor)
        # Ignore the testing savepoints
        if zsa_savepoints.state == 'begun' and check_constraints.state != 'checking':
            watcher.notice({
                "zsa_savepoints-state": zsa_savepoints.state,
                "check_constraints-state": check_constraints.state,
                "statement": statement,
                "parameters": parameters,
                "context": context,
                "executemany": executemany,
            })

    yield watcher

    sqlalchemy.event.remove(conn, 'after_cursor_execute', after_cursor_execute)


@pytest.yield_fixture
def no_deps(conn, DBSession):
    notice_pytest_fixtures(conn, DBSession)

    if NO_SERVER_FIXTURES:
        yield 'NO_SERVER_FIXTURES'
        return

    session = DBSession()

    @sqlalchemy.event.listens_for(session, 'after_flush')
    def check_dependencies(session, flush_context):
        ignored(session)
        assert not flush_context.cycles

    @sqlalchemy.event.listens_for(conn, "before_execute", retval=True)
    def before_execute(conn, clauseelement, multiparams, params):
        ignored(conn)
        return clauseelement, multiparams, params

    yield

    # TODO: At first glance, this cleanup looks wrong in a variety of ways.
    #       * Lack of try/finally
    #       * There are two defined .listens_for, but only one .remove
    #       * Not sure about what arguments remove takes, but is it supposed to match the .listens_for fns above?
    #       -kmp 7-Aug-2022
    sqlalchemy.event.remove(session, 'before_flush', check_dependencies)


@pytest.fixture(scope='session')
def wsgi_server_host_port():

    if NO_SERVER_FIXTURES:
        return 'NO_SERVER_FIXTURES'

    return webtest.http.get_free_port()


@pytest.fixture(scope='session')
def wsgi_server_app(app):
    notice_pytest_fixtures(app)

    if NO_SERVER_FIXTURES:
        return 'NO_SERVER_FIXTURES'

    return app


@pytest.mark.fixture_cost(100)
@pytest.yield_fixture(scope='session')
def wsgi_server(request, wsgi_server_app, wsgi_server_host_port):
    notice_pytest_fixtures(request, wsgi_server_app, wsgi_server_host_port)

    if NO_SERVER_FIXTURES:
        yield 'NO_SERVER_FIXTURES'
        return

    host, port = wsgi_server_host_port

    server = webtest.http.StopableWSGIServer.create(
        wsgi_server_app,
        host=host,
        port=port,
        threads=1,
        channel_timeout=60,
        cleanup_interval=10,
        expose_tracebacks=True,
        clear_untrusted_proxy_headers=True,
    )
    assert server.wait()

    yield 'http://%s:%s' % wsgi_server_host_port

    server.shutdown()



from base64 import (
    b64decode,
    b64encode,
)
from hashlib import sha384
from passlib.registry import register_crypt_handler
from passlib.utils import handlers as uh


def includeme(config):
    register_crypt_handler(SNOWHash)


class SNOWHash(uh.StaticHandler):
    """ a special snowflake of a  password hashing scheme

    Cryptographic strength of the hashing function is less of a concern for
    randomly generated passwords.
    """
    name = 'snowflake_hash'
    checksum_chars = uh.PADDED_BASE64_CHARS
    checksum_size = 64

    setting_kwds = ('salt_before', 'salt_after', 'salt_base')
    salt_before = b"186ED79BAEXzeusdioIsdklnw88e86cd73"
    salt_after = b"<*#$*(#)!DSDFOUIHLjksdf"
    salt_base = b64decode(b"""\
Kf8r/S37L/kh9yP1JfMn8TnvO+096z/pMecz5TXjN+EJ3wvdDdsP2QHXA9UF0wfRGc8bzR3LH8kR
xxPFFcMXwWm/a71tu2+5YbdjtWWzZ7F5r3utfat/qXGnc6V1o3ehSZ9LnU2bT5lBl0OVRZNHkVmP
W41di1+JUYdThVWDV4Gpf6t9rXuveaF3o3Wlc6dxuW+7bb1rv2mxZ7NltWO3YYlfi12NW49ZgVeD
VYVTh1GZT5tNnUufSZFHk0WVQ5dB6T/rPe077znhN+M15TPnMfkv+y39K/8p8SfzJfUj9yHJH8sd
zRvPGcEXwxXFE8cR2Q/bDd0L3wnRB9MF1QPXASn/K/0t+y/5Ifcj9SXzJ/E57zvtPes/6THnM+U1
4zfhCd8L3Q3bD9kB1wPVBdMH0RnPG80dyx/JEccTxRXDF8Fpv2u9bbtvuWG3Y7Vls2exea97rX2r
f6lxp3OldaN3oUmfS51Nm0+ZQZdDlUWTR5FZj1uNXYtfiVGHU4VVg1eBqX+rfa17r3mhd6N1pXOn
cblvu229a79psWezZbVjt2GJX4tdjVuPWYFXg1WFU4dRmU+bTZ1Ln0mRR5NFlUOXQek/6z3tO+85
4TfjNeUz5zH5L/st/Sv/KfEn8yX1I/chyR/LHc0bzxnBF8MVxRPHEdkP2w3dC98J0QfTBdUD1wE=
""")

    def _calc_checksum(self, secret):
        if not isinstance(secret, bytes):
            secret = secret.encode('utf-8')
        salted = self.salt_before + secret + self.salt_after + b'\0'
        if len(salted) > len(self.salt_base):
            raise ValueError("Password too long")
        salted += self.salt_base[len(salted):]
        chk = sha384(salted).digest()
        return b64encode(chk).decode("ascii")



import unittest

from pyramid.interfaces import IAuthenticationPolicy
from pyramid.security import Authenticated, Everyone
from pyramid.testing import DummyRequest
from zope.interface.verify import verifyObject, verifyClass
from .authentication import NamespacedAuthenticationPolicy


class TestNamespacedAuthenticationPolicy(unittest.TestCase):
    """ This is a modified version of TestRemoteUserAuthenticationPolicy
    """
    def _getTargetClass(self):
        return NamespacedAuthenticationPolicy

    def _makeOne(self, namespace='user',
                base='pyramid.authentication.RemoteUserAuthenticationPolicy',
                *args, **kw):
        return self._getTargetClass()(namespace, base, *args, **kw)

    def test_class_implements_IAuthenticationPolicy(self):
        klass = self._makeOne().__class__
        verifyClass(IAuthenticationPolicy, klass)

    def test_instance_implements_IAuthenticationPolicy(self):
        verifyObject(IAuthenticationPolicy, self._makeOne())

    def test_unauthenticated_userid_returns_None(self):
        request = DummyRequest(environ={})
        policy = self._makeOne()
        self.assertEqual(policy.unauthenticated_userid(request), None)

    def test_unauthenticated_userid(self):
        request = DummyRequest(environ={'REMOTE_USER':'fred'})
        policy = self._makeOne()
        self.assertEqual(policy.unauthenticated_userid(request), 'user.fred')

    def test_authenticated_userid_None(self):
        request = DummyRequest(environ={})
        policy = self._makeOne()
        self.assertEqual(policy.authenticated_userid(request), None)

    def test_authenticated_userid(self):
        request = DummyRequest(environ={'REMOTE_USER':'fred'})
        policy = self._makeOne()
        self.assertEqual(policy.authenticated_userid(request), 'user.fred')

    def test_effective_principals_None(self):
        request = DummyRequest(environ={})
        policy = self._makeOne()
        self.assertEqual(policy.effective_principals(request), [Everyone])

    def test_effective_principals(self):
        request = DummyRequest(environ={'REMOTE_USER':'fred'})
        policy = self._makeOne()
        self.assertEqual(policy.effective_principals(request),
                         [Everyone, Authenticated, 'user.fred'])

    def test_remember(self):
        request = DummyRequest(environ={'REMOTE_USER':'fred'})
        policy = self._makeOne()
        result = policy.remember(request, 'fred')
        self.assertEqual(result, [])

    def test_forget(self):
        request = DummyRequest(environ={'REMOTE_USER':'fred'})
        policy = self._makeOne()
        result = policy.forget(request)
        self.assertEqual(result, [])

    # From TestSessionAuthenticationPolicy

    def test_session_remember(self):
        request = DummyRequest()
        policy = self._makeOne(
                    base='pyramid.authentication.SessionAuthenticationPolicy',
                    prefix='')
        result = policy.remember(request, 'user.fred')
        self.assertEqual(request.session.get('userid'), 'fred')
        self.assertEqual(result, [])
        self.assertEqual(policy.unauthenticated_userid(request), 'user.fred')

    def test_session_forget(self):
        request = DummyRequest(session={'userid':'fred'})
        policy = self._makeOne(
                    base='pyramid.authentication.SessionAuthenticationPolicy',
                    prefix='')
        result = policy.forget(request)
        self.assertEqual(request.session.get('userid'), None)
        self.assertEqual(result, [])

    def test_session_forget_no_identity(self):
        request = DummyRequest()
        policy = self._makeOne(
                    base='pyramid.authentication.SessionAuthenticationPolicy',
                    prefix='')
        result = policy.forget(request)
        self.assertEqual(request.session.get('userid'), None)
        self.assertEqual(result, [])



from contextlib import contextmanager
from typing import Iterator, Optional, Union
from unittest import mock

import pytest
from dcicutils.item_model_utils import JsonObject
from dcicutils.testing_utils import patch_context
from pyramid.request import Request

from .. import item_models as item_models_module
from ..item_models import PortalItem


SOME_UUID = "uuid1234"
SOME_ITEM_PROPERTIES = {
    "uuid": SOME_UUID,
}
SOME_AUTH = {"key": "some_key", "secret": "some_secret"}


@contextmanager
def patch_from_identifier_and_auth(**kwargs) -> Iterator[mock.MagicMock]:
    with patch_context(
        item_models_module.PortalItem.from_identifier_and_auth, **kwargs
    ) as mock_item:
        yield mock_item


@contextmanager
def patch_from_identifier_and_request(**kwargs) -> Iterator[mock.MagicMock]:
    with patch_context(
        item_models_module.PortalItem.from_identifier_and_request, **kwargs
    ) as mock_item:
        yield mock_item


def mock_request(embed_exception: bool = False) -> mock.MagicMock:
    result = mock.create_autospec(Request, instance=True)
    embed_method = mock.MagicMock()
    if embed_exception:
        embed_method.side_effect = Exception
    result.embed = embed_method
    return result


def get_portal_item(
    auth: Optional[JsonObject] = SOME_AUTH,
    fetch_links: bool = False,
    request: Optional[Request] = None,
) -> PortalItem:
    return PortalItem({}, auth=auth, fetch_links=fetch_links, request=request)


class TestPortalItem:
    @pytest.mark.parametrize(
        "auth,request_param,exception_expected",
        [
            (None, None, True),
            (SOME_AUTH, None, False),
            (None, mock_request(), False),
            (SOME_AUTH, mock_request(), False),
        ],
    )
    def test_from_identifier_and_existing_item(
        self,
        auth: Union[JsonObject, None],
        request_param: Union[Request, None],
        exception_expected: bool,
    ) -> None:
        identifier = SOME_UUID
        fetch_links = True
        portal_item = get_portal_item(
            fetch_links=fetch_links, auth=auth, request=request_param
        )
        with patch_from_identifier_and_auth() as mock_from_identifier_and_auth:
            with patch_from_identifier_and_request() as mock_from_identifier_and_request:
                if exception_expected:
                    with pytest.raises(RuntimeError):
                        PortalItem.from_identifier_and_existing_item(
                            identifier, portal_item
                        )
                else:
                    result = PortalItem.from_identifier_and_existing_item(
                        identifier, portal_item
                    )
                    if auth:
                        assert result == mock_from_identifier_and_auth.return_value
                        mock_from_identifier_and_auth.assert_called_once_with(
                            identifier, auth, fetch_links=fetch_links
                        )
                        mock_from_identifier_and_request.assert_not_called()
                    elif request_param:
                        assert result == mock_from_identifier_and_request.return_value
                        mock_from_identifier_and_request.assert_called_once_with(
                            identifier, request_param, fetch_links=fetch_links
                        )
                        mock_from_identifier_and_auth.assert_not_called()

    @pytest.mark.parametrize("exception_expected", [True, False])
    def test_get_item_via_request(self, exception_expected: bool) -> None:
        identifier = SOME_UUID
        request = mock_request(embed_exception=exception_expected)
        result = PortalItem._get_item_via_request(identifier, request)
        request.embed.assert_called_once_with(identifier, "@@object")
        if exception_expected:
            assert result == {}
        else:
            assert result == request.embed.return_value



import pytest
import webtest
import boto3

from base64 import b64decode
from dcicutils.ff_utils import parse_s3_bucket_and_key_url
from moto import mock_s3
from unittest import mock
from .. import attachment as attachment_module
from ..attachment import file_type, guess_mime_type, system_mime_type, fallback_mime_type, DEFAULT_FALLBACK_MIME_TYPE


# Test for blob storage

RED_DOT = """data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAUA
AAAFCAYAAACNbyblAAAAHElEQVQI12P4//8/w38GIAXDIBKE0DHxgljNBAAO
9TXL0Y4OHwAAAABJRU5ErkJggg=="""

BLUE_DOT = """data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAA
oAAAAKAQMAAAC3/F3+AAAACXBIWXMAAA7DAAAOwwHHb6hkAA
AAAXNSR0IArs4c6QAAAARnQU1BAACxjwv8YQUAAAAgY0hSTQ
AAeiYAAICEAAD6AAAAgOgAAHUwAADqYAAAOpgAABdwnLpRPA
AAAANQTFRFALfvPEv6TAAAAAtJREFUCB1jYMAHAAAeAAEBGN
laAAAAAElFTkSuQmCC"""


@pytest.fixture
def testing_download(testapp):
    url = '/testing-downloads/'
    item = {
        'attachment': {
            'download': 'red-dot.png',
            'href': RED_DOT,
        },
        'attachment2': {
            'download': 'blue-dot.png',
            'href': BLUE_DOT,
        },
    }
    res = testapp.post_json(url, item, status=201)
    return res.location


class TestAttachment:
    """ Tests attachments with RDBBlobStorage """

    @staticmethod
    def test_download_create(testapp, testing_download):
        res = testapp.get(testing_download)
        attachment = res.json['attachment']
        attachment2 = res.json['attachment2']

        assert attachment['href'] == '@@download/attachment/red-dot.png'
        assert attachment['type'] == 'image/png'
        assert attachment['width'] == 5
        assert attachment['height'] == 5
        assert attachment['md5sum'] == 'b60ab2708daec7685f3d412a5e05191a'
        url = testing_download + '/' + attachment['href']
        res = testapp.get(url)
        assert res.content_type == 'image/png'
        assert res.body == b64decode(RED_DOT.split(',', 1)[1])

        assert attachment2['href'] == '@@download/attachment2/blue-dot.png'
        assert attachment2['type'] == 'image/png'
        assert attachment2['width'] == 10
        assert attachment2['height'] == 10
        assert attachment2['md5sum'] == '013f03aa088adb19aa226c3439bda179'
        url = testing_download + '/' + attachment2['href']
        res = testapp.get(url)
        assert res.content_type == 'image/png'
        assert res.body == b64decode(BLUE_DOT.split(',', 1)[1])

    @staticmethod
    def test_download_update(testapp, testing_download):
        item = {
            'attachment': {
                'download': 'blue-dot.png',
                'href': BLUE_DOT,
            },
            'attachment2': {
                'download': 'red-dot.png',
                'href': RED_DOT,
            },
        }
        testapp.put_json(testing_download, item, status=200)
        res = testapp.get(testing_download)
        attachment = res.json['attachment']
        attachment2 = res.json['attachment2']

        assert attachment['href'] == '@@download/attachment/blue-dot.png'
        url = testing_download + '/' + attachment['href']
        res = testapp.get(url)
        assert res.content_type == 'image/png'
        assert res.body == b64decode(BLUE_DOT.split(',', 1)[1])

        assert attachment2['href'] == '@@download/attachment2/red-dot.png'
        url = testing_download + '/' + attachment2['href']
        res = testapp.get(url)
        assert res.content_type == 'image/png'
        assert res.body == b64decode(RED_DOT.split(',', 1)[1])

    @staticmethod
    def test_download_update_no_change(testapp, testing_download):
        item = {
            'attachment': {
                'download': 'red-dot.png',
                'href': '@@download/attachment/red-dot.png',
            },
            'attachment2': {
                'download': 'blue-dot.png',
                'href': '@@download/attachment2/blue-dot.png',
            },
        }
        testapp.put_json(testing_download, item, status=200)

        res = testapp.get(testing_download)
        attachment = res.json['attachment']
        attachment2 = res.json['attachment2']
        assert attachment['href'] == '@@download/attachment/red-dot.png'
        assert attachment2['href'] == '@@download/attachment2/blue-dot.png'

    @staticmethod
    def test_download_update_one(testapp, testing_download):
        item = {
            'attachment': {
                'download': 'red-dot.png',
                'href': '@@download/attachment/red-dot.png',
            },
            'attachment2': {
                'download': 'red-dot.png',
                'href': RED_DOT,
            },
        }
        testapp.put_json(testing_download, item, status=200)

        res = testapp.get(testing_download)
        attachment = res.json['attachment']
        attachment2 = res.json['attachment2']

        assert attachment['href'] == '@@download/attachment/red-dot.png'
        url = testing_download + '/' + attachment['href']
        res = testapp.get(url)
        assert res.content_type == 'image/png'
        assert res.body == b64decode(RED_DOT.split(',', 1)[1])

        assert attachment2['href'] == '@@download/attachment2/red-dot.png'
        url = testing_download + '/' + attachment2['href']
        res = testapp.get(url)
        assert res.content_type == 'image/png'
        assert res.body == b64decode(RED_DOT.split(',', 1)[1])

    @staticmethod
    def test_download_update_with_premade_href(testapp):
        """
        Test the functionality of PATCHing the attachment information directly
        (which is usually generated by the propsheet)
        """
        post_res = testapp.post_json('/testing-downloads/', {}, status=201)
        item_uuid = post_res.json['@graph'][0]['uuid']
        res = testapp.get('/testing-downloads/' + item_uuid).follow()
        # downloading will not work!
        with pytest.raises(webtest.AppError) as excinfo:
            testapp.get('/testing-downloads/' + item_uuid + '/@@download/attachment/blue-dot.png')
        assert 'Cannot find downloads' in str(excinfo.value)

        assert 'attachment' not in res.json
        patch_body = {
            'attachment': {
                'download': 'blue-dot.png',
                'href': '@@download/attachment/blue-dot.png',
            }
        }
        res2 = testapp.patch_json('/testing-downloads/' + item_uuid, patch_body, status=200)
        attachment = res2.json['@graph'][0]['attachment']
        assert attachment['href'] == '@@download/attachment/blue-dot.png'

    @staticmethod
    def test_download_remove_one(testapp, testing_download):
        item = {
            'attachment': {
                'download': 'red-dot.png',
                'href': '@@download/attachment/red-dot.png',
            },
        }
        testapp.put_json(testing_download, item, status=200)

        res = testapp.get(testing_download)
        assert 'attachment' in res.json
        assert 'attachment2' not in res.json

        url = testing_download + '/@@download/attachment2/red-dot.png'
        testapp.get(url, status=404)

    @staticmethod
    @pytest.mark.parametrize(
        'href',
        [
            '@@download/attachment/another.png',
            'http://example.com/another.png',
        ])
    def test_download_update_bad_change(testapp, testing_download, href):
        item = {'attachment': {
            'download': 'red-dot.png',
            'href': href,
        }}
        testapp.put_json(testing_download, item, status=422)

    @staticmethod
    @pytest.mark.parametrize(
        'href',
        [
            'http://example.com/another.png',
            'data:image/png;base64,NOT_BASE64',
            'data:image/png;NOT_A_PNG',
            'data:text/plain;asdf',
        ])
    def test_download_create_bad_change(testapp, href):
        url = '/testing-downloads/'
        item = {'attachment': {
            'download': 'red-dot.png',
            'href': href,
        }}
        testapp.post_json(url, item, status=422)

    @staticmethod
    def test_download_create_wrong_extension(testapp):
        url = '/testing-downloads/'
        item = {'attachment': {
            'download': 'red-dot.jpg',
            'href': RED_DOT,
        }}
        testapp.post_json(url, item, status=422)

    @staticmethod
    def test_download_create_w_wrong_md5sum(testapp):
        url = '/testing-downloads/'
        item = {'attachment': {
            'download': 'red-dot.jpg',
            'href': RED_DOT,
            'md5sum': 'deadbeef',
        }}
        testapp.post_json(url, item, status=422)


@pytest.mark.filterwarnings('ignore:stream argument is deprecated')
class TestAttachmentEncrypted:
    """ TODO: It would be great if this class could be elegantly merged with the previous one.
        Note though that the tests do differ in some subtle ways. For example, when using
        s3blobs, we follow s3 redirects via pre-signed URLs - these URLs are parsed via
        helper function and acquired through boto3 in order to interact through moto. -Will Jan 4 2022
    """
    url = '/testing-downloads/'

    def testing_encrypted_download(self, testapp):
        item = {
            'attachment': {
                'download': 'red-dot.png',
                'href': RED_DOT,
            },
            'attachment2': {
                'download': 'blue-dot.png',
                'href': BLUE_DOT,
            },
        }
        res = testapp.post_json(self.url, item, status=201)
        return res.location

    def create_bucket_and_post_download(self, testapp):
        """ Bootstraps moto and creates some data to work with. """
        blob_bucket = 'encoded-4dn-blobs'  # note that this bucket exists but is mocked out here
        conn = boto3.resource('s3', region_name='us-east-1')
        conn.create_bucket(Bucket=blob_bucket)
        return self.testing_encrypted_download(testapp)

    @staticmethod
    def get_attachment_from_s3(client, url):
        """ Uses ff_utils.parse_s3_bucket_and_key_url to parse the s3 URL in our data into it's
            bucket, key pairs and acquires/reads the content.
        """
        bucket, key = parse_s3_bucket_and_key_url(url)
        return client.get_object(Bucket=bucket, Key=key)['Body'].read()

    def attachment_is_red_dot(self, client, url):
        """ Fails assertion if the url is not the RED_DOT """
        content = self.get_attachment_from_s3(client, url)
        assert content == b64decode(RED_DOT.split(',', 1)[1])

    def attachment_is_blue_dot(self, client, url):
        """ Fails assertion if the url is not the BLUE_DOT """
        content = self.get_attachment_from_s3(client, url)
        assert content == b64decode(BLUE_DOT.split(',', 1)[1])

    @mock_s3
    def test_download_create_encrypted(self, encrypted_testapp):
        loc = self.create_bucket_and_post_download(encrypted_testapp)
        res = encrypted_testapp.get(loc)
        attachment = res.json['attachment']
        attachment2 = res.json['attachment2']

        assert attachment['href'] == '@@download/attachment/red-dot.png'
        assert attachment['type'] == 'image/png'
        assert attachment['width'] == 5
        assert attachment['height'] == 5
        assert attachment['md5sum'] == 'b60ab2708daec7685f3d412a5e05191a'
        url = loc + '/' + attachment['href']

        res = encrypted_testapp.get(url)  # follow redirect to s3
        assert res.content_type == 'application/json'
        self.attachment_is_red_dot(boto3.client('s3'), res.location)

        assert attachment2['href'] == '@@download/attachment2/blue-dot.png'
        assert attachment2['type'] == 'image/png'
        assert attachment2['width'] == 10
        assert attachment2['height'] == 10
        assert attachment2['md5sum'] == '013f03aa088adb19aa226c3439bda179'
        url = loc + '/' + attachment2['href']
        res = encrypted_testapp.get(url)
        assert res.content_type == 'application/json'
        self.attachment_is_blue_dot(boto3.client('s3'), res.location)

    @mock_s3
    def test_download_update_encrypted(self, encrypted_testapp):
        item = {
            'attachment': {
                'download': 'blue-dot.png',
                'href': BLUE_DOT,
            },
            'attachment2': {
                'download': 'red-dot.png',
                'href': RED_DOT,
            },
        }
        loc = self.create_bucket_and_post_download(encrypted_testapp)
        encrypted_testapp.put_json(loc, item, status=200)
        res = encrypted_testapp.get(loc)
        attachment = res.json['attachment']
        attachment2 = res.json['attachment2']

        assert attachment['href'] == '@@download/attachment/blue-dot.png'
        url = loc + '/' + attachment['href']
        res = encrypted_testapp.get(url)
        self.attachment_is_blue_dot(boto3.client('s3'), res.location)

        assert attachment2['href'] == '@@download/attachment2/red-dot.png'
        url = loc + '/' + attachment2['href']
        res = encrypted_testapp.get(url)
        assert res.content_type == 'application/json'
        self.attachment_is_red_dot(boto3.client('s3'), res.location)

    @mock_s3
    def test_download_update_no_change_encrypted(self, encrypted_testapp):
        item = {
            'attachment': {
                'download': 'red-dot.png',
                'href': '@@download/attachment/red-dot.png',
            },
            'attachment2': {
                'download': 'blue-dot.png',
                'href': '@@download/attachment2/blue-dot.png',
            },
        }
        loc = self.create_bucket_and_post_download(encrypted_testapp)
        encrypted_testapp.put_json(loc, item, status=200)

        res = encrypted_testapp.get(loc)
        attachment = res.json['attachment']
        attachment2 = res.json['attachment2']
        assert attachment['href'] == '@@download/attachment/red-dot.png'
        assert attachment2['href'] == '@@download/attachment2/blue-dot.png'

    @mock_s3
    def test_download_update_one_encrypted(self, encrypted_testapp):
        item = {
            'attachment': {
                'download': 'red-dot.png',
                'href': '@@download/attachment/red-dot.png',
            },
            'attachment2': {
                'download': 'red-dot.png',
                'href': RED_DOT,
            },
        }
        loc = self.create_bucket_and_post_download(encrypted_testapp)
        encrypted_testapp.put_json(loc, item, status=200)

        res = encrypted_testapp.get(loc)
        attachment = res.json['attachment']
        attachment2 = res.json['attachment2']

        assert attachment['href'] == '@@download/attachment/red-dot.png'
        url = loc + '/' + attachment['href']
        res = encrypted_testapp.get(url)
        self.attachment_is_red_dot(boto3.client('s3'), res.location)

        assert attachment2['href'] == '@@download/attachment2/red-dot.png'
        url = loc + '/' + attachment2['href']
        res = encrypted_testapp.get(url)
        assert res.content_type == 'application/json'
        self.attachment_is_red_dot(boto3.client('s3'), res.location)

    @mock_s3
    def test_download_update_with_premade_href_encrypted(self, encrypted_testapp):
        """
        Test the functionality of PATCHing the attachment information directly
        (which is usually generated by the propsheet)
        """
        post_res = encrypted_testapp.post_json(self.url, {}, status=201)
        item_uuid = post_res.json['@graph'][0]['uuid']
        res = encrypted_testapp.get(self.url + item_uuid).follow()
        # downloading will not work!
        with pytest.raises(webtest.AppError) as excinfo:
            encrypted_testapp.get(self.url + item_uuid + '/@@download/attachment/blue-dot.png')
        assert 'Cannot find downloads' in str(excinfo.value)

        assert 'attachment' not in res.json
        patch_body = {
            'attachment': {
                'download': 'blue-dot.png',
                'href': '@@download/attachment/blue-dot.png',
            }
        }
        res2 = encrypted_testapp.patch_json(self.url + item_uuid, patch_body, status=200)
        attachment = res2.json['@graph'][0]['attachment']
        assert attachment['href'] == '@@download/attachment/blue-dot.png'

    @mock_s3
    def test_download_remove_one_encrypted(self, encrypted_testapp):
        item = {
            'attachment': {
                'download': 'red-dot.png',
                'href': '@@download/attachment/red-dot.png',
            },
        }
        loc = self.create_bucket_and_post_download(encrypted_testapp)
        encrypted_testapp.put_json(loc, item, status=200)

        res = encrypted_testapp.get(loc)
        assert 'attachment' in res.json
        assert 'attachment2' not in res.json

        url = loc + '/@@download/attachment2/red-dot.png'
        encrypted_testapp.get(url, status=404)

    @mock_s3
    @pytest.mark.parametrize(
        'href',
        [
            '@@download/attachment/another.png',
            'http://example.com/another.png',
        ])
    def test_download_update_bad_change_encrypted(self, encrypted_testapp, href):
        item = {'attachment': {
            'download': 'red-dot.png',
            'href': href,
        }}
        loc = self.create_bucket_and_post_download(encrypted_testapp)
        encrypted_testapp.put_json(loc, item, status=422)

    @mock_s3
    @pytest.mark.parametrize(
        'href',
        [
            'http://example.com/another.png',
            'data:image/png;base64,NOT_BASE64',
            'data:image/png;NOT_A_PNG',
            'data:text/plain;asdf',
        ])
    def test_download_create_bad_change_encrypted(self, encrypted_testapp, href):
        item = {'attachment': {
            'download': 'red-dot.png',
            'href': href,
        }}
        encrypted_testapp.post_json(self.url, item, status=422)

    @mock_s3
    def test_download_create_wrong_extension_encrypted(self, encrypted_testapp):
        item = {'attachment': {
            'download': 'red-dot.jpg',
            'href': RED_DOT,
        }}
        encrypted_testapp.post_json(self.url, item, status=422)

    @mock_s3
    def test_download_create_w_wrong_md5sum_encrypted(self, encrypted_testapp):
        item = {'attachment': {
            'download': 'red-dot.jpg',
            'href': RED_DOT,
            'md5sum': 'deadbeef',
        }}
        encrypted_testapp.post_json(self.url, item, status=422)


def test_file_type():

    assert file_type("/xyz/foo") == ""
    assert file_type("/xyz/foo.abc") == ".abc"
    assert file_type("/xyz/foo.abc.def") == ".def"

    assert file_type("xyz/foo.abc") == ".abc"
    assert file_type("foo.abc") == ".abc"


def test_system_mime_type():
    check_mime_type(system_mime_type, unknown=None)


def test_fallback_mime_type():
    check_mime_type(fallback_mime_type, unknown=None)


def test_guess_mime_type():
    check_mime_type(guess_mime_type, unknown=DEFAULT_FALLBACK_MIME_TYPE)
    with mock.patch.object(attachment_module, "system_mime_type") as mock_system_mime_type:
        mock_system_mime_type.return_value = None
        assert mock_system_mime_type.call_count == 0
        check_mime_type(guess_mime_type, unknown=DEFAULT_FALLBACK_MIME_TYPE)
        assert mock_system_mime_type.call_count > 0


def check_mime_type(guesser, *, unknown):
    assert guesser("/xyz/foo.txt") == "text/plain"
    assert guesser("/xyz/foo.text") == "text/plain"
    assert guesser("/xyz/foo.csv") == "text/csv"
    assert guesser("/xyz/foo.tsv") == "text/tab-separated-values"
    assert guesser("/xyz/foo.xlsx") == "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet"
    assert guesser("/xyz/foo.xls") == "application/vnd.ms-excel"
    assert guesser("/xyz/foo.docx") == "application/vnd.openxmlformats-officedocument.wordprocessingml.document"
    assert guesser("/xyz/foo.doc") == "application/msword"
    assert guesser("/xyz/foo.htm") == "text/html"
    assert guesser("/xyz/foo.html") == "text/html"
    assert guesser("/xyz/foo.json") == "application/json"

    # There is no recommende extension for "application/ld+json"
    # assert guesser(...) == "application/ld+json"

    # Random type just returns None
    assert guesser("/xyz/foo.xyzzy") is unknown



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, Optional, Union

import structlog
from pyramid.request import Request

from dcicutils.item_model_utils import (
    JsonObject,
    PortalItem as PortalItemWithoutRequests,
)


logger = structlog.getLogger(__name__)


@dataclass(frozen=True)
class PortalItem(PortalItemWithoutRequests):
    request: Optional[Request] = field(default=None, hash=False)

    def get_request(self) -> Union[Request, None]:
        return self.request

    @classmethod
    def from_properties(
        cls,
        properties: JsonObject,
        fetch_links=False,
        auth=None,
        request=None,
        **kwargs: Any,
    ) -> PortalItem:
        return cls(properties, fetch_links=fetch_links, auth=auth, request=request)

    @classmethod
    def from_identifier_and_existing_item(
        cls, identifier: str, existing_item: PortalItem, **kwargs: Any
    ) -> PortalItem:
        fetch_links = existing_item.should_fetch_links()
        auth = existing_item.get_auth()
        request = existing_item.get_request()
        if auth:
            return cls.from_identifier_and_auth(
                identifier, auth, fetch_links=fetch_links, **kwargs
            )
        if request:
            return cls.from_identifier_and_request(
                identifier, request, fetch_links=fetch_links, **kwargs
            )
        raise RuntimeError(
            "Unable to fetch given identifier without auth key or request"
        )

    @classmethod
    def from_properties_and_existing_item(
        cls, properties: JsonObject, existing_item: PortalItem, **kwargs: Any
    ) -> PortalItem:
        fetch_links = existing_item.should_fetch_links()
        auth = existing_item.get_auth()
        request = existing_item.get_request()
        return cls.from_properties(
            properties, fetch_links=fetch_links, auth=auth, request=request, **kwargs
        )

    @classmethod
    def from_identifier_and_request(
        cls, identifier: str, request: Request, fetch_links: bool, **kwargs: Any
    ) -> PortalItem:
        properties = cls._get_item_via_request(identifier, request)
        return cls.from_properties(
            properties, request=request, fetch_links=fetch_links, **kwargs
        )

    @classmethod
    def _get_item_via_request(cls, identifier: str, request: Request) -> JsonObject:
        try:
            result = request.embed(identifier, "@@object")
        except Exception as e:
            logger.exception(f"Unable to fetch identifer {identifier}: {e}")
            result = {}
        return result
